import numpy as np
from numba import njit, prange


def prox_21m(v, sigma, weight=1.0, axis=0):
    """
    Computes weighted version of

    prox_{sigma * || . ||_{21}}(v)

     Assumed that v has shape (nband, nbasis, nymax, nxmax) where

    nband   - number of imaging bands
    nbasis  - number of orthogonal bases
    nxmax   - number of x coefficients for each basis (must be equal)
    nymax   - number of y coefficients for each basis (must be equal)

    sigma  - scalar setting overall threshold level
    weight  - (nbasis, nymax, nxmax) setting relative weights per components in the
              "MFS" cube
    """
    l2_norm = np.sum(v, axis=axis)  # drops axis
    l2_soft = np.maximum(np.abs(l2_norm) - sigma * weight, 0.0) * np.sign(l2_norm)
    mask = l2_norm != 0
    ratio = np.zeros(mask.shape, dtype=v.dtype)
    ratio[mask] = l2_soft[mask] / l2_norm[mask]
    return v * np.expand_dims(ratio, axis=axis)  # restores axis


@njit(nogil=True, cache=True, parallel=True)
def prox_21m_numba(v, result, lam, sigma=1.0, weight=None):
    """
    Computes weighted version of

    prox_{sigma * || . ||_{21}}(v)

    Assumed that v has shape (nband, nbasis, nymax, nxmax) where

    nband   - number of imaging bands
    nbasis  - number of orthogonal bases
    nxmax   - number of x coefficients for each basis (must be equal)
    nymax   - number of y coefficients for each basis (must be equal)

    """
    nband, nbasis, nymax, nxmax = v.shape
    # sum over band axis upfront?
    # vsum = np.sum(v, axis=0)
    # result = np.zeros((nband, nbasis, ntot))
    for b in range(nbasis):
        vb = v[:, b]
        weightb = weight[b]
        resultb = result[:, b]
        for i in prange(nymax):
            for j in range(nxmax):
                vbisum = np.sum(vb[:, i, j])/sigma
                if not vbisum:
                    resultb[:, i, j] = 0.0
                    continue
                absvbi = np.abs(vbisum)
                softvbi = np.maximum(absvbi - lam*weightb[i, j]/sigma, 0.0) #* vbisum/absvbi
                resultb[:, i, j] = vb[:, i, j] * softvbi / absvbi /sigma


def dual_update(v, x, psiH, lam, sigma=1.0, weight=1.0):
    vp = v.copy()
    vout = np.zeros_like(v)
    psiH(x, vout)
    vtilde = vp + sigma * vout
    # return vtilde
    v = vtilde - sigma * prox_21m(vtilde/sigma, lam/sigma, weight=weight)
    return v



@njit(nogil=True, cache=True, parallel=True)
def dual_update_numba(vp, v, lam, sigma=1.0, weight=None):
    """
    Computes dual update

    Assumed that v has shape (nband, nbasis, nymax, nxmax) where

    nband   - number of imaging bands
    nbasis  - number of orthogonal bases
    nxmax   - number of x coefficients for each basis (must be equal)
    nymax   - number of y coefficients for each basis (must be equal)

    v is initialised with psiH(xp) and will be updated
    """
    nband, nbasis, nymax, nxmax = v.shape
    for b in range(nbasis):
        # select out basis
        # vtildeb = vp[:, b] + sigma * v[:, b]
        # weightb = weight[b]
        for i in prange(nymax):  # without the prange it segfaults when parallel=True?
            vtildebi = vp[:, b, i] + sigma * v[:, b, i]
            weightbi = weight[b, i]
            for j in range(nxmax):
                vtildebij = vtildebi[:, j]
                absvbijsum = np.abs(np.sum(vtildebij)/sigma)  # sum over band axis
                v[:, b, i, j] = vtildebij
                if absvbijsum:
                    softvbij = np.maximum(absvbijsum - lam*weightbi[j]/sigma, 0.0)
                    v[:, b, i, j] *= (1 - softvbij / absvbijsum)


@njit(nogil=True, cache=True, parallel=True)
def dual_update_numba_dist(vp, v, lam, sigma=1.0, weight=None):
    """
    Computes dual update

    Assumed that v has shape (nband, nbasis, nymax, nxmax) where

    nband   - number of imaging bands
    nbasis  - number of orthogonal bases
    nxmax   - number of x coefficients for each basis (must be equal)
    nymax   - number of y coefficients for each basis (must be equal)

    v is initialised with psiH(xp) and will be updated
    """
    nband, nbasis, nymax, nxmax = v.shape
    for b in range(nbasis):
        # select out basis
        # vtildeb = vp[:, b] + sigma * v[:, b]
        # weightb = weight[b]
        for i in prange(nymax):  # WTF without the prange it segfaults when parallel=True
            vtildebi = vp[:, b, i] + sigma * v[:, b, i]
            weightbi = weight[b, i]
            for j in range(nxmax):
                vtildebij = vtildebi[:, j]
                absvbijsum = np.abs(np.sum(vtildebij)/sigma)  # sum over band axis
                v[:, b, i, j] = vtildebij
                if absvbijsum:
                    softvbij = np.maximum(absvbijsum - lam*weightbi[j]/sigma, 0.0)
                    v[:, b, i, j] *= (1-softvbij / absvbijsum)



"""
Dask wrappers around the wgridder. These operators are per band
because we are not guaranteed that each imaging band has the same
number of rows after BDA.
"""
import concurrent.futures as cf
from time import time
import numpy as np
import numba
from numba import njit, prange, literally, types
from numba.extending import overload
import concurrent.futures as cf
import xarray as xr
import dask.array as da
from ducc0.wgridder.experimental import vis2dirty, dirty2vis
from ducc0.fft import r2c
from ducc0.misc import resize_thread_pool
from pfb.utils.weighting import counts_to_weights, _compute_counts
from pfb.utils.beam import eval_beam
from pfb.utils.naming import xds_from_list
from pfb.utils.misc import fitcleanbeam
from pfb.utils.stokes import stokes_funcs
from pfb.utils.misc import JIT_OPTIONS, _es_kernel
from scipy.constants import c as lightspeed
iFs = np.fft.ifftshift
Fs = np.fft.fftshift


def wgridder_conventions(l0, m0):
    '''
    Returns

    flip_u, flip_v, flip_w, x0, y0

    according to the conventions documented here https://github.com/mreineck/ducc/issues/34

    Note that these conventions are stored as dataset attributes in order
    to call the operators acting on datasets with a consistent convention.
    '''
    return False, True, False, -l0, -m0


def vis2im(uvw,
           freq,
           vis,
           wgt,
           mask,
           nx, ny,
           cellx, celly,
           l0, m0,
           epsilon,
           precision,
           do_wgridding,
           divide_by_n,
           nthreads,
           sigma_min, sigma_max,
           double_precision_accumulation):
    uvw = np.require(uvw, dtype=np.float64)
    freq = np.require(freq, np.float64)

    if precision.lower() == 'single':
        real_type = np.float32
        complex_type = np.complex64
    elif precision.lower() == 'double':
        real_type = np.float64
        complex_type = np.complex128

    vis = np.require(vis, dtype=complex_type)

    if wgt is not None:
        wgt = np.require(wgt, dtype=real_type)

    if mask is not None:
        mask = np.require(mask, dtype=np.uint8)

    flip_u, flip_v, flip_w, x0, y0 = wgridder_conventions(l0, m0)

    return vis2dirty(uvw=uvw,
                     freq=freq,
                     vis=vis,
                     wgt=wgt,
                     mask=mask,
                     npix_x=nx, npix_y=ny,
                     pixsize_x=cellx, pixsize_y=celly,
                     center_x=x0, center_y=y0,
                     epsilon=epsilon,
                     flip_u=flip_u,
                     flip_v=flip_v,
                     flip_w=flip_w,
                     do_wgridding=do_wgridding,
                     divide_by_n=divide_by_n,
                     nthreads=nthreads,
                     sigma_min=sigma_min, sigma_max=sigma_max,
                     double_precision_accumulation=double_precision_accumulation)


def im2vis(uvw,
           freq,
           image,
           cellx,
           celly,
           freq_bin_idx,
           freq_bin_counts,
           l0=0, m0=0,
           epsilon=1e-7,
           do_wgridding=True,
           divide_by_n=False,
           nthreads=1):
    # adjust for chunking
    # need a copy here if using multiple row chunks
    freq_bin_idx2 = freq_bin_idx - freq_bin_idx.min()
    flip_u, flip_v, flip_w, x0, y0 = wgridder_conventions(l0, m0)
    nband, nx, ny = image.shape
    nrow = uvw.shape[0]
    nchan = freq.size
    vis = np.zeros((nrow, nchan), dtype=np.result_type(image, np.complex64))
    for i in range(nband):
        ind = slice(freq_bin_idx2[i], freq_bin_idx2[i] + freq_bin_counts[i])
        vis[:, ind] = dirty2vis(
            uvw=uvw,
            freq=freq[ind],
            dirty=image[i],
            pixsize_x=cellx,
            pixsize_y=celly,
            center_x=x0,
            center_y=y0,
            flip_u=flip_u,
            flip_v=flip_v,
            flip_w=flip_w,
            epsilon=epsilon,
            nthreads=nthreads,
            do_wgridding=do_wgridding,
            divide_by_n=divide_by_n)
    return vis


# we still need the collections interface here for the degridder
def comps2vis(
            uvw,
            utime,
            freq,
            rbin_idx, rbin_cnts,
            tbin_idx, tbin_cnts,
            fbin_idx, fbin_cnts,
            region_mask,
            mds,
            modelf,
            tfunc,
            ffunc,
            epsilon=1e-7,
            nthreads=1,
            do_wgridding=True,
            divide_by_n=False,
            freq_min=-np.inf,
            freq_max=np.inf,
            product='I'):

    # determine output type
    complex_type = da.result_type(mds.coefficients.dtype, np.complex64)
    ncorr_out = len(product)

    return da.blockwise(_comps2vis, 'rfc',
                        uvw, 'r3',
                        utime, 'r',
                        freq, 'f',
                        rbin_idx, 'r',
                        rbin_cnts, 'r',
                        tbin_idx, 'r',
                        tbin_cnts, 'r',
                        fbin_idx, 'f',
                        fbin_cnts, 'f',
                        region_mask, None,
                        mds, None,
                        modelf, None,
                        tfunc, None,
                        ffunc, None,
                        epsilon, None,
                        nthreads, None,
                        do_wgridding, None,
                        divide_by_n, None,
                        freq_min, None,
                        freq_max, None,
                        product, None,
                        new_axes={'c': ncorr_out},
                        # it should be getting these from uvw and freq?
                        adjust_chunks={'r': uvw.chunks[0]},
                        dtype=complex_type,
                        align_arrays=False)


def _comps2vis(
            uvw,
            utime,
            freq,
            rbin_idx, rbin_cnts,
            tbin_idx, tbin_cnts,
            fbin_idx, fbin_cnts,
            region_mask,
            mds,
            modelf,
            tfunc,
            ffunc,
            epsilon=1e-7,
            nthreads=1,
            do_wgridding=True,
            divide_by_n=False,
            freq_min=-np.inf,
            freq_max=np.inf,
            product='I'):
    return _comps2vis_impl(
                        uvw[0],
                        utime,
                        freq,
                        rbin_idx, rbin_cnts,
                        tbin_idx, tbin_cnts,
                        fbin_idx, fbin_cnts,
                        region_mask,
                        mds,
                        modelf,
                        tfunc,
                        ffunc,
                        epsilon=epsilon,
                        nthreads=nthreads,
                        do_wgridding=do_wgridding,
                        divide_by_n=divide_by_n,
                        freq_min=freq_min,
                        freq_max=freq_max,
                        product=product)



def _comps2vis_impl(uvw,
                    utime,
                    freq,
                    rbin_idx, rbin_cnts,
                    tbin_idx, tbin_cnts,
                    fbin_idx, fbin_cnts,
                    region_mask,
                    mds,
                    modelf,
                    tfunc,
                    ffunc,
                    epsilon=1e-7,
                    nthreads=1,
                    do_wgridding=True,
                    divide_by_n=False,
                    freq_min=-np.inf,
                    freq_max=np.inf,
                    product='I'):
    # why is this necessary?
    resize_thread_pool(nthreads)

    # adjust for chunking
    # need a copy here if using multiple row chunks
    rbin_idx2 = rbin_idx - rbin_idx.min()
    tbin_idx2 = tbin_idx - tbin_idx.min()
    fbin_idx2 = fbin_idx - fbin_idx.min()

    ntime = tbin_idx.size
    nband = fbin_idx.size

    nrow = uvw.shape[0]
    nchan = freq.size
    nstokes_out = len(product)
    vis = np.zeros((nrow, nchan, nstokes_out),
                   dtype=np.result_type(mds.coefficients.dtype, np.complex64))
    if not ((freq>=freq_min) & (freq<=freq_max)).any():
        return vis

    comps = mds.coefficients.values
    Ix = mds.location_x.values
    Iy = mds.location_y.values
    cellx = mds.cell_rad_x
    celly = mds.cell_rad_x
    nx = mds.npix_x
    ny = mds.npix_y
    # these are taken from dataset attrs to make sure they remain consistent
    x0 = mds.center_x
    y0 = mds.center_y
    flip_u = mds.flip_u
    flip_v = mds.flip_v
    flip_w = mds.flip_w
    for t in range(ntime):
        indt = slice(tbin_idx2[t], tbin_idx2[t] + tbin_cnts[t])
        # TODO - clean up this logic. row_mapping holds the number of rows per
        # utime and there are multiple utimes per row chunk
        indr = slice(rbin_idx2[indt][0], rbin_idx2[indt][-1] + rbin_cnts[indt][-1])
        for b in range(nband):
            indf = slice(fbin_idx2[b], fbin_idx2[b] + fbin_cnts[b])
            f = freq[indf]
            # don't degrid outside requested frequency range
            if not ((f>=freq_min) & (f<=freq_max)).any():
                continue
            # render components to image
            tout = tfunc(np.mean(utime[indt]))
            fout = ffunc(np.mean(freq[indf]))
            image = np.zeros((nx, ny), dtype=comps.dtype)
            image[Ix, Iy] = modelf(tout, fout, *comps[:, :])  # too magical?
            if np.any(region_mask):
                image = np.where(region_mask, image, 0.0)
                for c in range(nstokes_out):
                    vis[indr, indf, c] = dirty2vis(uvw=uvw,
                                        freq=f,
                                        dirty=image,
                                        pixsize_x=cellx, pixsize_y=celly,
                                        center_x=x0, center_y=y0,
                                        flip_u=flip_u,
                                        flip_v=flip_v,
                                        flip_w=flip_w,
                                        epsilon=epsilon,
                                        do_wgridding=do_wgridding,
                                        divide_by_n=divide_by_n,
                                        nthreads=nthreads)

    return vis


def image_data_products(dsl,
                        dsp,
                        nx, ny,
                        nx_psf, ny_psf,
                        cellx, celly,
                        output_name,
                        attrs,
                        model=None,
                        robustness=None,
                        l0=0.0, m0=0.0,
                        nthreads=1,
                        epsilon=1e-7,
                        do_wgridding=True,
                        double_accum=True,
                        l2_reweight_dof=None,
                        do_dirty=True,
                        do_psf=True,
                        do_residual=True,
                        do_weight=True,
                        do_noise=False,
                        do_beam=False,
                        min_padding=1.7):
    '''
    Function to compute image space data products in one go

        dirty
        psf
        residual
        noise
        beam
        imweight
        wsum

    Assumes all datasets are concatenatable and will compute weighted
    sum of beam
    '''
    resize_thread_pool(nthreads)
    flip_u, flip_v, flip_w, x0, y0 = wgridder_conventions(l0, m0)
    signu = -1.0 if flip_u else 1.0
    signv = -1.0 if flip_v else 1.0
    # we need these in the test because of flipped wgridder convention
    # https://github.com/mreineck/ducc/issues/34
    signx = -1.0 if flip_u else 1.0
    signy = -1.0 if flip_v else 1.0

    # TODO - assign ug,vg-coordinates
    x = (-nx/2 + np.arange(nx)) * cellx + x0
    y = (-ny/2 + np.arange(ny)) * celly + y0
    n = np.sqrt(1 - x0**2 - y0**2)

    # expects a list
    if isinstance(dsl, str):
        dsl = [dsl]

    dsl = xds_from_list(dsl, nthreads=nthreads)

    ncorr = dsl[0].corr.size

    # need to take weighted sum of beam before concat
    beam = np.zeros((ncorr, nx, ny), dtype=float)
    wsumb = np.zeros(ncorr, dtype=float)
    freq = dsl[0].FREQ.values  # must all be the same
    xx, yy = np.meshgrid(np.rad2deg(x), np.rad2deg(y), indexing='ij')
    for i, ds in enumerate(dsl):
        wgt = ds.WEIGHT.values
        mask = ds.MASK.values
        for c in range(ncorr):
            wsumt = (wgt[c]*mask).sum()
            wsumb[c] += wsumt
            l_beam = ds.l_beam.values
            m_beam = ds.m_beam.values
            beamt = eval_beam(ds.BEAM.values[c], l_beam, m_beam,
                              xx, yy)
            beam[c] += beamt * wsumt
            assert (ds.FREQ.values == freq).all()
        ds = ds.drop_vars(('BEAM','FREQ'))
        dsl[i] = ds.drop_dims(('l_beam', 'm_beam'))

    # TODO - weighted sum of beam computed using natural weights?
    beam /= wsumb[:, None, None]

    ds = xr.concat(dsl, dim='row')
    uvw = ds.UVW.values
    vis = ds.VIS.values
    wgt = ds.WEIGHT.values
    mask = ds.MASK.values
    bandid = int(ds.bandid)

    _, nrow, nchan = vis.shape

    coords = {
        'x': x,
        'y': y,
        'corr': ds.corr.values,
        'bpar': ['BMAJ', 'BMIN', 'BPA']
    }

    # output ds
    dso = xr.Dataset(attrs=attrs, coords=coords)
    dso['FREQ'] = (('chan',), freq)

    if model is None:
        if l2_reweight_dof:
            raise ValueError('Requested l2 reweight but no model passed in. '
                             'Perhaps transfer model from somewhere?')
    else:
        # do not apply weights in this direction
        # actually model vis, this saves memory
        residual_vis = np.zeros_like(vis)
        for c in range(ncorr):
            dirty2vis(
                uvw=uvw,
                freq=freq,
                dirty=model[c],
                pixsize_x=cellx,
                pixsize_y=celly,
                center_x=x0,
                center_y=y0,
                epsilon=epsilon,
                do_wgridding=do_wgridding,
                flip_u=flip_u,
                flip_v=flip_v,
                flip_w=flip_w,
                nthreads=nthreads,
                divide_by_n=False,  # incorporate in smooth beam
                sigma_min=1.1, sigma_max=3.0,
                vis=residual_vis[c])

        # this is an attempt to avoid the tmp model_vis array
        residual_vis *= -1  # negate model
        residual_vis += vis

    if l2_reweight_dof:
        if dsp:
            dsp = xds_from_list([dsp], drop_all_but='WEIGHT')
            wgtp = dsp[0].WEIGHT.values
        else:
            wgtp = 1.0
        # mask needs to be bool here
        ressq = (residual_vis*wgtp*residual_vis.conj()).real
        # we are currently doing this per correlation
        # should it be done jointly?
        ssq = ressq[:, mask>0].sum(axis=-1)
        ovar = ssq/mask.sum()  # same mask for all corrs
        if ovar:
            # scale the natural weights
            # RHS is weight relative to unity since wgtp included in ressq
            # meani = np.mean(ressq[:, mask>0]/ovar)
            # stdi = np.std(ressq[:, mask>0]/ovar)
            # print(f"Band {bandid} before: mean = {meani:.3e}, std = {stdi:.3e}")
            denom = (l2_reweight_dof + ressq/ovar[:, None, None])
            # the expectation value of the complex ressq above is 2
            # if we do this jointly over correlations this should be 2*ncorr
            wgt *= (l2_reweight_dof + 2)/denom
        else:
            wgt = None
    # re-evaluate since robustness and or wgt after reweight may change
    if robustness is not None:
        numba_threads = np.maximum(nthreads, 1)
        numba.set_num_threads(numba_threads)
        # we need to compute the weights on the padded grid
        # but we don't have control over the optimal gridding
        # parameters so assume a minimum
        nx_pad = int(np.ceil(min_padding*nx))
        if nx_pad%2:
            nx_pad += 1
        ny_pad = int(np.ceil(min_padding*ny))
        if ny_pad%2:
            ny_pad += 1
        counts = _compute_counts(uvw,
                                 freq,
                                 mask,
                                 wgt,
                                 nx_pad, ny_pad,
                                 cellx, celly,
                                 uvw.dtype,
                                 k=0,
                                 # limit number of grids
                                 ngrid=np.minimum(numba_threads, 8),
                                 usign=1.0 if flip_u else -1.0,
                                 vsign=1.0 if flip_v else -1.0)
        wgt = counts_to_weights(
            counts,
            uvw,
            freq,
            wgt,
            mask,
            nx_pad, ny_pad,
            cellx, celly,
            robustness,
            usign=1.0 if flip_u else -1.0,
            vsign=1.0 if flip_v else -1.0)

    # these are always used together
    if do_weight:
        dso['WEIGHT'] = (('corr', 'row','chan'), wgt)
        dso['UVW'] = (('row', 'three'), uvw)
        dso['MASK'] = (('row','chan'), mask)

    wsum = wgt[:, mask.astype(bool)].sum(axis=-1)
    dso['WSUM'] = (('corr',), wsum)

    if do_dirty:
        dirty = np.zeros((ncorr, nx, ny), dtype=float)
        for c in range(ncorr):
            vis2dirty(
                uvw=uvw,
                freq=freq,
                vis=vis[c],
                wgt=wgt[c],
                mask=mask,
                npix_x=nx, npix_y=ny,
                pixsize_x=cellx, pixsize_y=celly,
                center_x=x0, center_y=y0,
                epsilon=epsilon,
                flip_u=flip_u,
                flip_v=flip_v,
                flip_w=flip_w,
                do_wgridding=do_wgridding,
                divide_by_n=False,  # incorporte in smooth beam
                nthreads=nthreads,
                sigma_min=1.1, sigma_max=3.0,
                double_precision_accumulation=double_accum,
                dirty=dirty[c])
        dso['DIRTY'] = (('corr', 'x','y'), dirty)

    if do_psf:
        if x0 or y0:
            freqfactor = 2j*np.pi*freq[None, :]/lightspeed
            psf_vis = np.exp(freqfactor*(signu*uvw[:, 0:1]*x0*signx +
                             signv*uvw[:, 1:2]*y0*signy -
                             uvw[:, 2:]*(n-1)))

        else:
            nrow, _ = uvw.shape
            nchan = freq.size
            tmp = np.ones((1,), dtype=vis.dtype)
            # should be tiny
            psf_vis = np.broadcast_to(tmp, (nrow, nchan))

        psf = np.zeros((ncorr, nx_psf, ny_psf), dtype=float)
        for c in range(ncorr):
            vis2dirty(
                uvw=uvw,
                freq=freq,
                vis=psf_vis,
                wgt=wgt[c],
                mask=mask,
                npix_x=nx_psf, npix_y=ny_psf,
                pixsize_x=cellx, pixsize_y=celly,
                center_x=x0, center_y=y0,
                flip_u=flip_u,
                flip_v=flip_v,
                flip_w=flip_w,
                epsilon=epsilon,
                do_wgridding=do_wgridding,
                divide_by_n=False,  # incorporte in smooth beam
                nthreads=nthreads,
                sigma_min=1.1, sigma_max=3.0,
                double_precision_accumulation=double_accum,
                dirty=psf[c])

        # get FT of psf
        psfhat = r2c(iFs(psf, axes=(1, 2)), axes=(1, 2),
                     nthreads=nthreads,
                     forward=True, inorm=0)

        dso["PSF"] = (('corr', 'x_psf', 'y_psf'), psf)
        dso["PSFHAT"] = (('corr', 'x_psf', 'yo2'), psfhat)

        # add natural resolution info
        # normalised internally
        gausspar = fitcleanbeam(psf, level=0.5, pixsize=1.0)
        dso["PSFPARSN"] = (('corr', 'bpar'), np.array(gausspar))


    if do_residual and model is not None:
        residual = np.zeros((ncorr, nx, ny), dtype=float)
        for c in range(ncorr):
            vis2dirty(
                uvw=uvw,
                freq=freq,
                vis=residual_vis[c],
                wgt=wgt[c],
                mask=mask,
                npix_x=nx, npix_y=ny,
                pixsize_x=cellx, pixsize_y=celly,
                center_x=x0, center_y=y0,
                flip_u=flip_u,
                flip_v=flip_v,
                flip_w=flip_w,
                epsilon=epsilon,
                do_wgridding=do_wgridding,
                divide_by_n=False,  # incorporte in smooth beam
                nthreads=nthreads,
                sigma_min=1.1, sigma_max=3.0,
                double_precision_accumulation=double_accum,
                dirty=residual[c])

        dso['MODEL'] = (('corr','x','y'), model)
        dso['RESIDUAL'] = (('corr','x','y'), residual)

    if do_noise:
        # sample noise and project into image space
        _, nrow, nchan = vis.shape
        from pfb.utils.misc import parallel_standard_normal
        noise = np.zeros((ncorr, nx, ny), dtype=float)
        for c in range(ncorr):
            vis = (parallel_standard_normal((nrow, nchan)) +
                1j*parallel_standard_normal((nrow, nchan)))
            wmask = wgt[c] > 0.0
            vis[wmask] /= np.sqrt(wgt[c, wmask])
            vis[~wmask] = 0j
            vis2dirty(
                uvw=uvw,
                freq=freq,
                vis=vis,
                wgt=wgt[c],
                mask=mask,
                npix_x=nx, npix_y=ny,
                pixsize_x=cellx, pixsize_y=celly,
                center_x=x0, center_y=y0,
                flip_u=flip_u,
                flip_v=flip_v,
                flip_w=flip_w,
                epsilon=epsilon,
                do_wgridding=do_wgridding,
                divide_by_n=False,  # incorporte in smooth beam
                nthreads=nthreads,
                sigma_min=1.1, sigma_max=3.0,
                double_precision_accumulation=double_accum,
                dirty=noise[c])

        dso['NOISE'] = (('corr', 'x','y'), noise)

    if do_beam:
        if beam is not None:
            dso['BEAM'] = (('corr', 'x', 'y'), beam)
        else:
            dso['BEAM'] = (('corr', 'x', 'y'), np.ones((ncorr, nx, ny),
                                                       dtype=wgt.dtype))

    # save
    dso = dso.assign_attrs(wsum=wsum, x0=x0, y0=y0, l0=l0, m0=m0,
                           flip_u=flip_u,
                           flip_v=flip_v,
                           flip_w=flip_w)
    dso.to_zarr(output_name, mode='a')

    outputs = {}
    # return residual to report stats
    if do_residual and model is not None:
        outputs['residual'] = residual
    else:
        outputs['residual'] = dirty
    # PSF returned to fit MFS PSFPARS
    if do_psf:
        outputs['psf'] = psf
    outputs['wsum'] = wsum
    outputs['timeid'] = attrs['timeid']
    return outputs


def compute_residual(dsl,
                     nx, ny,
                     cellx, celly,
                     output_name,
                     model,
                     nthreads=1,
                     epsilon=1e-7,
                     do_wgridding=True,
                     double_accum=True,
                     verbosity=1):
    '''
    Function to compute residual and write it to disk
    '''
    resize_thread_pool(nthreads)
    # expects a list
    if isinstance(dsl, str):
        dsl = [dsl]

    tii = time()

    # currently only a single dds
    ti = time()
    ds = xds_from_list(dsl, nthreads=nthreads)[0]

    uvw = ds.UVW.values
    wgt = ds.WEIGHT.values
    mask = ds.MASK.values
    beam = ds.BEAM.values
    dirty = ds.DIRTY.values
    freq = ds.FREQ.values
    flip_u = ds.flip_u
    flip_v = ds.flip_v
    flip_w = ds.flip_w
    x0 = ds.x0
    y0 = ds.y0
    ncorr = ds.corr.size

    tread = time() - ti

    ti = time()
    # do not apply weights in this direction
    convim = np.zeros_like(dirty)
    for c in range(ncorr):
        model_vis = dirty2vis(
            uvw=uvw,
            freq=freq,
            dirty=beam[c]*model[c],
            pixsize_x=cellx,
            pixsize_y=celly,
            center_x=x0,
            center_y=y0,
            flip_u=flip_u,
            flip_v=flip_v,
            flip_w=flip_w,
            epsilon=epsilon,
            do_wgridding=do_wgridding,
            nthreads=nthreads,
            divide_by_n=False,  # incorporate in smooth beam
            sigma_min=1.1, sigma_max=3.0,
            verbosity=0)
    
        vis2dirty(
            uvw=uvw,
            freq=freq,
            vis=model_vis,
            wgt=wgt[c],
            mask=mask,
            npix_x=nx, npix_y=ny,
            pixsize_x=cellx, pixsize_y=celly,
            center_x=x0, center_y=y0,
            flip_u=flip_u,
            flip_v=flip_v,
            flip_w=flip_w,
            epsilon=epsilon,
            do_wgridding=do_wgridding,
            divide_by_n=False,  # incorporate in smooth beam
            nthreads=nthreads,
            sigma_min=1.1, sigma_max=3.0,
            double_precision_accumulation=double_accum,
            verbosity=0,
            dirty=convim[c])
    tgrid = time() - ti

    # this is the once attenuated residual since
    # dirty is only attenuated once
    ti = time()
    residual = dirty - convim
    tdiff = time() - ti

    ti = time()
    ds['MODEL'] = (('corr', 'x','y'), model)
    ds['RESIDUAL'] = (('corr', 'x','y'), residual)
    tassign = time() - ti

    # we only need to write MODEL and RESIDUAL
    ds = ds[['RESIDUAL','MODEL']]

    # save
    # LB - Why is twrite still siginificant?
    ti = time()
    with cf.ThreadPoolExecutor(max_workers=1) as executor:
        future = executor.submit(ds.to_zarr, output_name, mode='a')
    twrite = time() - ti

    ttot = time() - tii
    ttally = tread + tgrid + tdiff + tassign + twrite
    if verbosity > 1:
        print(f'tread = {tread/ttot}')
        print(f'tgrid = {tgrid/ttot}')
        print(f'tdiff = {tdiff/ttot}')
        print(f'tassign = {tassign/ttot}')
        print(f'twrite = {twrite/ttot}')
        print(f'ttally = {ttally/ttot}')
    return residual, future

def dataset_to_zarr(ds, output_name):
    ds.to_zarr(output_name, mode='a')



# flake8: noqa
from contextlib import ExitStack
import click
from omegaconf import OmegaConf
from pfb.utils import logging as pfb_logging
from scabha.schema_utils import clickify_parameters
from pfb.parser.schemas import schema

log = pfb_logging.get_logger('MODEL2COMPS')


@click.command(context_settings={'show_default': True})
@clickify_parameters(schema.model2comps)
def model2comps(**kw):
    '''
    Convert model in dds to components.
    '''
    opts = OmegaConf.create(kw)

    from pfb.utils.naming import set_output_names
    opts, basedir, oname = set_output_names(opts)

    import psutil
    nthreads = psutil.cpu_count(logical=True)
    ncpu = psutil.cpu_count(logical=False)
    if opts.nthreads is None:
        opts.nthreads = nthreads//2
        ncpu = ncpu//2

    OmegaConf.set_struct(opts, True)

    from pfb import set_envs
    from ducc0.misc import resize_thread_pool, thread_pool_size
    resize_thread_pool(opts.nthreads)
    set_envs(opts.nthreads, ncpu)

    import time
    timestamp = time.strftime("%Y%m%d-%H%M%S")
    logname = f'{str(opts.log_directory)}/model2comps_{timestamp}.log'
    pfb_logging.log_to_file(logname)
    log.info(f'Logs will be written to {logname}')

    pfb_logging.log_options_dict(log, opts)

    with ExitStack() as stack:
        ti = time.time()
        if opts.from_fits:
            _model2comps_fits(**opts)
        else:
            _model2comps(**opts)


    log.info(f"All done after {time.time() - ti}s")

def _model2comps(**kw):
    opts = OmegaConf.create(kw)
    OmegaConf.set_struct(opts, True)

    import os
    import numpy as np
    from pfb.utils.naming import xds_from_url, xds_from_list
    from africanus.constants import c as lightspeed
    from pfb.utils.modelspec import fit_image_cube, eval_coeffs_to_slice
    from pfb.utils.fits import set_wcs, save_fits
    from pfb.utils.misc import norm_diff
    import xarray as xr
    import fsspec as fs
    from daskms.fsspec_store import DaskMSStore
    import json
    from casacore.quanta import quantity
    from pfb.utils.modelspec import eval_coeffs_to_slice

    basename = opts.output_filename
    if opts.fits_output_folder is not None:
        fits_oname = opts.fits_output_folder + '/' + basename.split('/')[-1]
    else:
        fits_oname = basename

    dds_name = f'{basename}_{opts.suffix}.dds'
    dds, dds_list = xds_from_url(dds_name)

    if opts.model_out is not None:
        coeff_name = opts.model_out
        fits_name = opts.model_out.rstrip('.mds') + '.fits'
    else:
        coeff_name = f'{basename}_{opts.suffix}_{opts.model_name.lower()}.mds'
        fits_name = f'{basename}_{opts.suffix}_{opts.model_name.lower()}.fits'

    mdsstore = DaskMSStore(coeff_name)
    if mdsstore.exists():
        if opts.overwrite:
            log.info(f"Overwriting {coeff_name}")
            mdsstore.rm(recursive=True)
        else:
            log.error_and_raise(f"{coeff_name} exists. Set --overwrite to overwrite it. ",
                                ValueError)


    cell_rad = dds[0].cell_rad
    cell_deg = np.rad2deg(cell_rad)

    # get cube info
    nband = 0
    ntime = 0
    mfreqs = []
    mtimes = []
    for ds in dds:
        mfreqs.append(ds.freq_out)
        mtimes.append(ds.time_out)
    mfreqs = np.unique(np.array(mfreqs))
    mtimes = np.unique(np.array(mtimes))
    nband = mfreqs.size
    ntime = mtimes.size

    # stack cube
    nx = dds[0].x.size
    ny = dds[0].y.size
    x0 = dds[0].x0
    y0 = dds[0].y0

    model = np.zeros((ntime, nband, nx, ny), dtype=np.float64)
    wsums = np.zeros((ntime, nband), dtype=np.float64)
    for ds in dds:
        b = int(ds.bandid)
        t = int(ds.timeid)
        model[t, b] = getattr(ds, opts.model_name).values
        wsums[t, b] += ds.WSUM.values[0]

    if not opts.use_wsum:
        wsums[...] = 1.0
    else:
        # normalise so ridge param has more intuitive meaning
        wsums /= wsums.max()

    # render input model onto model grid
    # TODO - this is not perfectly flux conservative
    if opts.mds is not None:
        mdsi_store = DaskMSStore(opts.mds)
        mdsi = xr.open_zarr(mdsi_store.url)
        # we only want to load these once
        coeffsi = mdsi.coefficients.values
        locxi = mdsi.location_x.values
        locyi = mdsi.location_y.values
        paramsi = mdsi.params.values
        for t, mtime in enumerate(mtimes):
            for b, mfreq in enumerate(mfreqs):
                model[t, b] += eval_coeffs_to_slice(
                    mtime,
                    mfreq,
                    coeffsi,
                    locxi, locyi,
                    mdsi.parametrisation,
                    paramsi,
                    mdsi.texpr,
                    mdsi.fexpr,
                    mdsi.npix_x, mdsi.npix_y,
                    mdsi.cell_rad_x, mdsi.cell_rad_y,
                    mdsi.center_x, mdsi.center_y,
                    nx, ny,
                    cell_rad, cell_rad,
                    x0, y0
                )

    if not np.any(model):
        log.error_and_raise(f'Model is empty', RuntimeError)
    radec = (dds[0].ra, dds[0].dec)

    if opts.out_freqs is not None:
        flow, fhigh, step = list(map(float, opts.out_freqs.split(':')))
        fsel = np.all(wsums > 0, axis=0)
        if flow < mfreqs.min():
            log.info(f"Linearly extrapolating to {flow:.3e}Hz")
            # linear extrapolation from first two non-null bands
            Ilow = np.argmax(fsel)  # first non-null index
            Ihigh = Ilow + 1
            while not fsel[Ihigh]:
                Ihigh += 1
            nudiff = mfreqs[Ihigh] - mfreqs[Ilow]
            slopes = (model[:, Ihigh] - model[:, Ilow])/nudiff
            intercepts = model[:, Ihigh] - slopes * mfreqs[Ihigh]
            mlow = slopes * flow + intercepts
            model = np.concatenate((mlow[:, None], model), axis=1)
            mfreqs = np.concatenate((np.array((flow,)), mfreqs))
            # TODO - duplicate first non-null value?
            wsums = np.concatenate((0.5*wsums[:, Ilow][:, None], wsums),
                                   axis=1)
            nband = mfreqs.size
        if fhigh > mfreqs.max():
            log.info(f"Linearly extrapolating to {fhigh:.3e}Hz")
            # linear extrapolation from last two non-null bands
            Ihigh = nband - np.argmax(fsel[::-1]) - 1  # last non-null index
            Ilow = Ihigh - 1
            while not fsel[Ilow]:
                Ilow -= 1
            nudiff = mfreqs[Ihigh] - mfreqs[Ilow]
            slopes = (model[:, Ihigh] - model[:, Ilow])/nudiff
            intercepts = model[:, Ihigh] - slopes * mfreqs[Ihigh]
            mhigh = slopes * fhigh + intercepts
            model = np.concatenate((model, mhigh[:, None]), axis=1)
            mfreqs = np.concatenate((mfreqs, np.array((fhigh,))))
            wsums = np.concatenate((wsums, 0.5*wsums[:, Ihigh][:, None]),
                                   axis=1)
            nband = mfreqs.size

    if opts.min_val is not None:
        model = np.where(model > opts.min_val, model, 0.0)

    if not np.any(model):
        log.error_and_raise(f'Model has no components above {opts.min_val}',
                            RuntimeError)

    if opts.nbasisf is None:
        nbasisf = nband-1
    else:
        nbasisf = opts.nbasisf

    nbasis = nbasisf
    log.info(f"Fitting {nband} bands with {nbasis} basis functions")
    try:
        coeffs, Ix, Iy, expr, params, texpr, fexpr = \
            fit_image_cube(mtimes, mfreqs, model,
                           wgt=wsums,
                           nbasisf=nbasisf,
                           method=opts.fit_mode,
                           sigmasq=opts.sigmasq)
    except np.linalg.LinAlgError as e:
        log.error_and_raise(f"Exception raised during fit ."
              f"Do you perhaps have empty sub-bands?"
              f"Decreasing nbasisf", e)

    # save interpolated dataset
    data_vars = {
        'coefficients': (('par', 'comps'), coeffs),
    }
    coords = {
        'location_x': (('x',), Ix),
        'location_y': (('y',), Iy),
        # 'shape_x':,
        'params': (('par',), params),  # already converted to list
        'times': (('t',), mtimes),  # to allow rendering to original grid
        'freqs': (('f',), mfreqs)
    }
    attrs = {
        'spec': 'genesis',
        'cell_rad_x': cell_rad,
        'cell_rad_y': cell_rad,
        'npix_x': nx,
        'npix_y': ny,
        'texpr': texpr,
        'fexpr': fexpr,
        'center_x': x0,
        'center_y': y0,
        'flip_u': dds[0].flip_u,
        'flip_v': dds[0].flip_v,
        'flip_w': dds[0].flip_w,
        'ra': dds[0].ra,
        'dec': dds[0].dec,
        'stokes': opts.product,  # I,Q,U,V, IQ/IV, IQUV
        'parametrisation': expr  # already converted to str
    }


    coeff_dataset = xr.Dataset(data_vars=data_vars,
                               coords=coords,
                               attrs=attrs)
    log.info(f'Writing interpolated model to {coeff_name}')

    if opts.out_format == 'zarr':
        coeff_dataset.to_zarr(mdsstore.url)
    elif opts.out_format == 'json':
        coeff_dict = coeff_dataset.to_dict()
        with fs.open(mdsstore.url, 'w') as f:
            json.dump(coeff_dict, f)


    # interpolation error
    modelo = np.zeros((nband, nx, ny))
    for b in range(nband):
        modelo[b] = eval_coeffs_to_slice(mtimes[0],
                                         mfreqs[b],
                                         coeffs,
                                         Ix, Iy,
                                         expr,
                                         params,
                                         texpr,
                                         fexpr,
                                         nx, ny,
                                         cell_rad, cell_rad,
                                         x0, y0,
                                         nx, ny,
                                         cell_rad, cell_rad,
                                         x0, y0)

    eps = norm_diff(modelo, model[0])
    log.info(f"Fractional interpolation error is {eps:.3e}")


    # TODO - doesn't work with multiple fields
    # render model to cube
    if opts.out_freqs is not None:
        flow, fhigh, step = list(map(float, opts.out_freqs.split(':')))
        nbando = int((fhigh - flow)/step)
        log.info(f"Rendering model cube to {nbando} output bands")
        freq_out = np.linspace(flow, fhigh, nbando)
        ra = dds[0].ra
        dec  = dds[0].dec
        hdr = set_wcs(cell_deg, cell_deg, nx, ny, [ra, dec],
                    freq_out, GuassPar=(1, 1, 0),  # fake for now
                    ms_time=mtimes[0])
        modelo = np.zeros((nbando, nx, ny))
        for b in range(nbando):
            modelo[b] = eval_coeffs_to_slice(mtimes[0],
                                            freq_out[b],
                                            coeffs,
                                            Ix, Iy,
                                            expr,
                                            params,
                                            texpr,
                                            fexpr,
                                            nx, ny,
                                            cell_rad, cell_rad,
                                            x0, y0,
                                            nx, ny,
                                            cell_rad, cell_rad,
                                            x0, y0)

        save_fits(modelo,
                  fits_name,
                  hdr,
                  overwrite=True)

def _model2comps_fits(**kw):
    opts = OmegaConf.create(kw)
    OmegaConf.set_struct(opts, True)

    import os
    import numpy as np
    from africanus.constants import c as lightspeed
    from pfb.utils.modelspec import fit_image_cube
    from pfb.utils.fits import set_wcs, save_fits, load_fits
    from pfb.utils.misc import eval_coeffs_to_slice, norm_diff
    import xarray as xr
    import fsspec as fs
    from daskms.fsspec_store import DaskMSStore
    import json
    from casacore.quanta import quantity
    from pfb.utils.misc import eval_coeffs_to_slice
    from astropy.io import fits
    from glob import glob

    basename = opts.output_filename
    if opts.fits_output_folder is not None:
        fits_oname = opts.fits_output_folder + '/' + basename.split('/')[-1]
    else:
        fits_oname = basename

    if opts.model_out is not None:
        coeff_name = opts.model_out
        fits_name = opts.model_out.rstrip('.mds') + '.fits'
    else:
        coeff_name = f'{basename}_{opts.suffix}_{opts.model_name.lower()}.mds'
        fits_name = f'{basename}_{opts.suffix}_{opts.model_name.lower()}.fits'

    mdsstore = DaskMSStore(coeff_name)
    if mdsstore.exists():
        if opts.overwrite:
            log.info(f"Overwriting {coeff_name}")
            mdsstore.rm(recursive=True)
        else:
            log.error_and_raise(f"{coeff_name} exists. "
                                "Set --overwrite to overwrite it. ",
                                RuntimeError)


    images_list = sorted(
            glob(f"{opts.from_fits}-[0-9][0-9][0-9][0-9]-model.fits"),
            key=os.path.getctime)
    if len(images_list) == 0:
        log.error_and_raise(f"No images found at {opts.from_fits}",
                            ValueError)

    # get cube info
    nband = len(images_list)
    ntime = 1
    mfreqs = []
    model_list = []
    cellx_deg = None
    celly_deg = None
    nx = None
    ny = None
    ra = None
    dec = None
    wsums = []
    for image in images_list:
        log.info(f"Loading {image}")
        hdu = fits.open(image)
        model_list.append(hdu[0].data.squeeze().T)
        mfreqs.append(hdu[0].header['CRVAL3'])
        if 'WSCVWSUM' in hdu[0].header:
            wsums.append(hdu[0].header['WSCVWSUM'])
        elif 'WSCIMGWG' in hdu[0].header:
            wsums.append(hdu[0].header['WSCIMGWG'])
        else:
            wsums.append(1.0)
        if cellx_deg is None:
            cellx_deg = np.abs(hdu[0].header['CDELT2'])
        else:
            assert cellx_deg == np.abs(hdu[0].header['CDELT2'])
        if celly_deg is None:
            celly_deg = np.abs(hdu[0].header['CDELT1'])
        else:
            assert celly_deg == np.abs(hdu[0].header['CDELT1'])
        if ny is None:
            ny = hdu[0].header['NAXIS2']
        else:
            assert ny == hdu[0].header['NAXIS2']
        if nx is None:
            nx = hdu[0].header['NAXIS1']
        else:
            assert nx == hdu[0].header['NAXIS1']
        if ra is None:
            ra = hdu[0].header['CRVAL1']
        else:
            assert ra == hdu[0].header['CRVAL1']
        if dec is None:
            dec = hdu[0].header['CRVAL2']
        else:
            assert dec == hdu[0].header['CRVAL2']
        hdu.close()
    mfreqs = np.unique(np.array(mfreqs))
    assert mfreqs.size == nband
    for b in range(nband):
        log.info(f"Found {mfreqs[b]}Hz model at index {b}")
    assert cellx_deg == celly_deg
    cell_deg = cellx_deg
    cell_rad = np.deg2rad(cell_deg)
    x0 = 0.0
    y0 = 0.0
    wsums = np.array(wsums, dtype=np.float64)[None, :]
    ra = np.deg2rad(ra)
    dec = np.deg2rad(dec)
    mtimes = np.ones((1,), dtype=np.float64)
    model = np.zeros((ntime, nband, nx, ny), dtype=np.float64)
    for b in range(nband):
        model[0, b] = model_list[b]

    if not opts.use_wsum:
        wsums[...] = 1.0
    else:
        # normalise so ridge param has more intuitive meaning
        wsums /= wsums.max()

    if not np.any(model):
        log.error_and_raise('Model is empty', ValueError)
    radec = (ra, dec)

    if opts.out_freqs is not None:
        flow, fhigh, step = list(map(float, opts.out_freqs.split(':')))
        fsel = np.all(wsums > 0, axis=0)
        if flow < mfreqs.min():
            log.info(f"Linearly extrapolating to {flow:.3e}Hz")
            # linear extrapolation from first two non-null bands
            Ilow = np.argmax(fsel)  # first non-null index
            Ihigh = Ilow + 1
            while not fsel[Ihigh]:
                Ihigh += 1
            nudiff = mfreqs[Ihigh] - mfreqs[Ilow]
            slopes = (model[:, Ihigh] - model[:, Ilow])/nudiff
            intercepts = model[:, Ihigh] - slopes * mfreqs[Ihigh]
            mlow = slopes * flow + intercepts
            model = np.concatenate((mlow[:, None], model), axis=1)
            mfreqs = np.concatenate((np.array((flow,)), mfreqs))
            # TODO - duplicate first non-null value?
            wsums = np.concatenate((0.5*wsums[:, Ilow][:, None], wsums),
                                   axis=1)
            nband = mfreqs.size
        if fhigh > mfreqs.max():
            log.info(f"Linearly extrapolating to {fhigh:.3e}Hz")
            # linear extrapolation from last two non-null bands
            Ihigh = nband - np.argmax(fsel[::-1]) - 1  # last non-null index
            Ilow = Ihigh - 1
            while not fsel[Ilow]:
                Ilow -= 1
            nudiff = mfreqs[Ihigh] - mfreqs[Ilow]
            slopes = (model[:, Ihigh] - model[:, Ilow])/nudiff
            intercepts = model[:, Ihigh] - slopes * mfreqs[Ihigh]
            mhigh = slopes * fhigh + intercepts
            model = np.concatenate((model, mhigh[:, None]), axis=1)
            mfreqs = np.concatenate((mfreqs, np.array((fhigh,))))
            wsums = np.concatenate((wsums, 0.5*wsums[:, Ihigh][:, None]),
                                   axis=1)
            nband = mfreqs.size

    if opts.min_val is not None:
        model = np.where(model > opts.min_val, model, 0.0)

    if not np.any(model):
        log.error_and_raise(f'Model has no components above {opts.min_val}',
                            ValueError)

    if opts.nbasisf is None:
        nbasisf = nband-1
    else:
        nbasisf = opts.nbasisf

    nbasis = nbasisf
    log.info(f"Fitting {nband} bands with {nbasis} basis functions")
    try:
        coeffs, Ix, Iy, expr, params, texpr, fexpr = \
            fit_image_cube(mtimes, mfreqs, model,
                           wgt=wsums,
                           nbasisf=nbasisf,
                           method=opts.fit_mode,
                           sigmasq=opts.sigmasq)
    except np.linalg.LinAlgError as e:
        log.error_and_raise(f"Exception raised during fit ."
              f"Do you perhaps have empty sub-bands?"
              f"Decreasing nbasisf", e)

    # save interpolated dataset
    data_vars = {
        'coefficients': (('par', 'comps'), coeffs),
    }
    coords = {
        'location_x': (('x',), Ix),
        'location_y': (('y',), Iy),
        # 'shape_x':,
        'params': (('par',), params),  # already converted to list
        'times': (('t',), mtimes),  # to allow rendering to original grid
        'freqs': (('f',), mfreqs)
    }
    attrs = {
        'spec': 'genesis',
        'cell_rad_x': cell_rad,
        'cell_rad_y': cell_rad,
        'npix_x': nx,
        'npix_y': ny,
        'texpr': texpr,
        'fexpr': fexpr,
        'center_x': x0,
        'center_y': y0,
        'flip_u': False,
        'flip_v': True,
        'flip_w': False,
        'ra': ra,
        'dec': dec,
        'stokes': opts.product,  # I,Q,U,V, IQ/IV, IQUV
        'parametrisation': expr  # already converted to str
    }


    coeff_dataset = xr.Dataset(data_vars=data_vars,
                               coords=coords,
                               attrs=attrs)
    log.info(f'Writing interpolated model to {coeff_name}')

    if opts.out_format == 'zarr':
        coeff_dataset.to_zarr(mdsstore.url)
    elif opts.out_format == 'json':
        coeff_dict = coeff_dataset.to_dict()
        with fs.open(mdsstore.url, 'w') as f:
            json.dump(coeff_dict, f)


    # interpolation error
    modelo = np.zeros((nband, nx, ny))
    for b in range(nband):
        modelo[b] = eval_coeffs_to_slice(mtimes[0],
                                         mfreqs[b],
                                         coeffs,
                                         Ix, Iy,
                                         expr,
                                         params,
                                         texpr,
                                         fexpr,
                                         nx, ny,
                                         cell_rad, cell_rad,
                                         x0, y0,
                                         nx, ny,
                                         cell_rad, cell_rad,
                                         x0, y0)

    eps = norm_diff(modelo, model[0])
    log.info(f"Fractional interpolation error is {eps:.3e}")


    # TODO - doesn't work with multiple fields
    # render model to cube
    if opts.out_freqs is not None:
        flow, fhigh, step = list(map(float, opts.out_freqs.split(':')))
        nbando = int((fhigh - flow)/step)
        log.info(f"Rendering model cube to {nbando} output bands")
        freq_out = np.linspace(flow, fhigh, nbando)
        hdr = set_wcs(cell_deg, cell_deg, nx, ny, [ra, dec],
                      freq_out, GuassPar=(1, 1, 0),  # fake for now
                      ms_time=mtimes[0])
        modelo = np.zeros((nbando, nx, ny))
        for b in range(nbando):
            modelo[b] = eval_coeffs_to_slice(mtimes[0],
                                            freq_out[b],
                                            coeffs,
                                            Ix, Iy,
                                            expr,
                                            params,
                                            texpr,
                                            fexpr,
                                            nx, ny,
                                            cell_rad, cell_rad,
                                            x0, y0,
                                            nx, ny,
                                            cell_rad, cell_rad,
                                            x0, y0)
    else:  # we just write it back at input frequencies as a sanity check
        hdr = set_wcs(cell_deg, cell_deg, nx, ny, [ra, dec],
                      mfreqs, GuassPar=(1, 1, 0), casambm=False)

    save_fits(modelo[:, None, :, :],
              fits_name,
              hdr,
              overwrite=True)



import pywt
import numpy as np
import numba
from numba import prange
from numba.typed import List

PARALLEL = True

@numba.njit(nogil=True, cache=True)
def coeff_size(nsignal, nfilter):
    return (nsignal + nfilter - 1)//2


@numba.njit(nogil=True, cache=True)
def coeff_size_at_level(nsignal, nfilter, level):
    ncoeff = nsignal

    for _ in range(level):
        ncoeff = (ncoeff + nfilter - 1)//2

    return ncoeff

@numba.njit(nogil=True, cache=True)
def get_buffer_blocks(image_shape, filter_size, level):

    result = List()

    dim0, dim1 = image_shape
    result.append(dim0 * dim1)

    for _ in range(level):
        dim0 = coeff_size(dim0, filter_size)
        dim1 = coeff_size(dim1, filter_size)

        result.pop()
        result.extend([dim0 * dim1] * 4)

    return result


@numba.njit(nogil=True, cache=True)
def quadrants_from_dwt_buffer(buffer, image_shape, filter_size, level):

    image_size0, image_size1 = image_shape

    blocks = get_buffer_blocks(image_shape, filter_size, level)

    approx_size0 = coeff_size_at_level(image_size0, filter_size, level)
    approx_size1 = coeff_size_at_level(image_size1, filter_size, level)
    approx_shape = (approx_size0, approx_size1)

    q11 = buffer[sum(blocks[:-4]): sum(blocks[:-3])].reshape(approx_shape)
    q10 = buffer[sum(blocks[:-3]): sum(blocks[:-2])].reshape(approx_shape)
    q01 = buffer[sum(blocks[:-2]): sum(blocks[:-1])].reshape(approx_shape)
    q00 = buffer[sum(blocks[:-1]): sum(blocks)].reshape(approx_shape)

    return (q11, q10, q01, q00)


@numba.njit(nogil=True, cache=True)
def get_idwt_buffer_blocks(image_shape, filter_size, level):

    result = List()

    dim0, dim1 = image_shape
    result.append(dim0 * dim1)

    for _ in range(level):
        dim0 = coeff_size(dim0, filter_size)
        dim1 = coeff_size(dim1, filter_size)

        result.pop()
        result.extend([dim0 * dim1] * 4)

    result[-1] = get_signal_shape_at_level(image_shape, filter_size, level)

    return result


@numba.njit(nogil=True, cache=True)
def quadrants_from_idwt_buffer(buffer, image_shape, filter_size, level):

    image_size0, image_size1 = image_shape

    blocks = get_idwt_buffer_blocks(image_shape, filter_size, level)

    signal_shape = signal_size_at_level(image_shape, filter_size, level)
    block_shape = get_block_shape_at_level(image.shape, rec_lo.size, i + 1)

    q11 = buffer[sum(blocks[:-4]): sum(blocks[:-3])].reshape(block_shape)
    q10 = buffer[sum(blocks[:-3]): sum(blocks[:-2])].reshape(block_shape)
    q01 = buffer[sum(blocks[:-2]): sum(blocks[:-1])].reshape(block_shape)
    q00 = buffer[sum(blocks[:-1]): sum(blocks)].reshape(signal_shape)

    return (q11, q10, q01, q00)


@numba.njit(nogil=True, cache=True)
def get_block_shape_at_level(image_shape, filter_size, level):

    dim0, dim1 = image_shape

    for _ in range(level):
        dim0 = coeff_size(dim0, filter_size)
        dim1 = coeff_size(dim1, filter_size)

    return (dim0, dim1)


@numba.njit(nogil=True, cache=True)
def get_signal_shape_at_level(image_shape, filter_size, level):

    dim0, dim1 = image_shape

    for _ in range(level):
        dim0 = coeff_size(dim0, filter_size)
        dim1 = coeff_size(dim1, filter_size)

    dim0 = signal_size(dim0, filter_size)
    dim1 = signal_size(dim1, filter_size)

    return (dim0, dim1)


@numba.njit(nogil=True, cache=True)
def get_buffer_size(image_shape, filter_size, level):
    return sum(get_buffer_blocks(image_shape, filter_size, level))

@numba.njit(nogil=True, cache=True)
def signal_size(ncoeff, nfilter):
    return 2*ncoeff - nfilter + 2

@numba.njit(nogil=True, cache=True)
def signal_size_at_level(ncoeff, nfilter, lvel):
    return 2*ncoeff - nfilter + 2


@numba.njit(nogil=True, cache=True, inline='always')
def downsampling_convolution(input, output, filter, step):

    i = step - 1
    o = 0
    N = input.shape[0]
    F = filter.shape[0]


    # left boundary overhang
    while i < F and i < N:
        fsum = input.dtype.type(0)  # init to zero with correct type
        j = 0

        while j <= i:
            fsum += filter[j] * input[i - j]
            j += 1

        output[o] = fsum

        i += step
        o += 1

    # center (if input equal or wider than filter: N >= F)
    while i < N:
        fsum = input.dtype.type(0)
        j = 0

        while j < F:
            fsum += input[i - j] * filter[j]
            j += 1

        output[o] = fsum

        i += step
        o += 1

    # center (if filter is wider than input: F > N)
    while i < F:
        fsum = input.dtype.type(0)
        j = 0
        j = i - N + 1

        while j <= i:
            fsum += filter[j] * input[i - j]
            j += 1

        output[o] = fsum

        i += step
        o += 1

    # right boundary overhang
    while i < N + F - 1:
        fsum = input.dtype.type(0)
        j = 0
        j = i - N + 1

        while j < F:
            fsum += filter[j] * input[i - j]
            j += 1

        output[o] = fsum

        i += step
        o += 1



@numba.njit(nogil=True, cache=True, parallel=PARALLEL)
def dwt2d_level(image, quadrants, dec_lo, dec_hi):
    """
    This function is not meant to be called directly
    """
    nx, ny = image.shape

    x_dim = coeff_size(nx, dec_lo.size)
    y_dim = coeff_size(ny, dec_lo.size)

    approx_x = np.zeros((y_dim, nx))
    detail_x = np.zeros((y_dim, nx))

    quadrant00 = quadrants[-1]
    quadrant01 = quadrants[-2]
    quadrant10 = quadrants[-3]
    quadrant11 = quadrants[-4]

    for i in prange(nx):
        # approx - low pass
        downsampling_convolution(image[i, :], approx_x[:, i], dec_lo, 2)
        # detail - high pass
        downsampling_convolution(image[i, :], detail_x[:, i], dec_hi, 2)

    for i in prange(y_dim):
        # high pass - high pass
        downsampling_convolution(detail_x[i, :], quadrant11[:, i], dec_hi, 2)
        # high pass - low pass
        downsampling_convolution(detail_x[i, :], quadrant01[:, i], dec_lo, 2)
        # low pass - high pass
        downsampling_convolution(approx_x[i, :], quadrant10[:, i], dec_hi, 2)
        # low pass - low pass
        downsampling_convolution(approx_x[i, :], quadrant00[:, i], dec_lo, 2)


@numba.njit(nogil=True, cache=True, parallel=PARALLEL)
def dwt2d(image, buffer, dec_lo, dec_hi, nlevel):
    '''
    image       - (nx, ny) signal
    buffer      - coeff buffer,the size of which is detemined by the level
                  and filter size
    dec_lo      - low pass decomposition filter
    dec_hi      - high pass decomposition filter
    '''

    image_shape = image.shape
    filter_size = dec_lo.size

    # Possibly unecessary?
    buffer[:image.size] = image.ravel()
    image = buffer[:image.size].reshape(image_shape)

    for level in range(1, nlevel + 1):

        quadrants = quadrants_from_dwt_buffer(
            buffer, image_shape, filter_size, level
        )

        dwt2d_level(image, quadrants, dec_lo, dec_hi)

        image = quadrants[-1]



@numba.njit(nogil=True, cache=True, inline='always')
def upsampling_convolution_valid_sf(input, filter, output):

    N = input.shape[0]
    F = filter.shape[0]
    O = output.shape[0]

    # if F % 2:
    #     raise ValueError("even filter required for upsampling")

    # if N < (F // 2):
    #     raise ValueError("input.shape[0] < (filter.shape[0] // 2)")

    o = 0
    i = (F // 2) - 1
    stopping_criteria = F // 2

    while i < N and o < O:
        sum_even = input.dtype.type(0)
        sum_odd = input.dtype.type(0)
        j = 0
        j2 = 0

        while j < stopping_criteria:
            input_element = input[i - j]
            sum_even += filter[j2] * input_element
            sum_odd += filter[j2 + 1] * input_element

            j += 1
            j2 += 2

        output[o] += sum_even
        output[o + 1] += sum_odd

        i += 1
        o += 2


@numba.njit(nogil=True, cache=True, parallel=PARALLEL)
def idwt2d_level(recon, quadrants, rec_lo, rec_hi):
    '''
    outT is a buffer the size of out.T
    tmp is a buffer the size of alpha
    tmpT is a buffer the size of alpha.T
    '''

    dec11, dec10, dec01, dec00 = quadrants

    quad_x_dim, quad_y_dim = dec11.shape  # Coefficient quadrant array shape

    rec_x_dim = signal_size(quad_x_dim, rec_lo.size)
    rec_y_dim = signal_size(quad_y_dim, rec_lo.size)

    approx_x = np.zeros((rec_x_dim, quad_y_dim))
    detail_x = np.zeros((rec_x_dim, quad_y_dim))

    for i in prange(quad_y_dim): # Upsampling convolution along column.
        upsampling_convolution_valid_sf(dec00[:, i], rec_lo, approx_x[:, i])
        upsampling_convolution_valid_sf(dec10[:, i], rec_hi, approx_x[:, i])
        upsampling_convolution_valid_sf(dec01[:, i], rec_lo, detail_x[:, i])
        upsampling_convolution_valid_sf(dec11[:, i], rec_hi, detail_x[:, i])

    recon[:] = 0  # Clear the buffer as we need to accumulate!

    for i in prange(rec_x_dim): # Upsampling convolution along column.
        upsampling_convolution_valid_sf(approx_x[i, :], rec_lo, recon[i, :])
        upsampling_convolution_valid_sf(detail_x[i, :], rec_hi, recon[i, :])


@numba.njit(nogil=True, cache=True, parallel=PARALLEL)
def idwt2d(buffer, image, rec_lo, rec_hi, nlevel):
    '''

    coeffs           - (nax, nay) coeffs
    image             - (nx, ny) output array
    imageT            - (ny, nx) signal buffer to avoid non-contiguous access
    tmpa            - (nax, nay) coeff buffer that avoids overwriting coeffs
    tmp             - (nay, nax) coeff buffer that avoids copy and transpose
    tmpT            - (nay, nax) coeff buffer that avoids copy and transpose
    ix              - dict keyed on level containing x start and stop indices i.e. ix[level] = (ix, fx)
    iy              - dict keyed on level containing y start and stop indices i.e. iy[level] = (iy, fy)
    '''

    quadrants = quadrants_from_dwt_buffer(
        buffer, image.shape, rec_lo.size, nlevel
    )

    blocks = get_buffer_blocks(image.shape, rec_lo.size, nlevel)
    block_dim = get_block_shape_at_level(image.shape, rec_lo.size, nlevel)
    recon_size = block_dim[0] * block_dim[1]
    recon_shape = (block_dim[0], block_dim[1])

    for i in range(nlevel - 1, -1, -1):

        block_dim = get_block_shape_at_level(image.shape, rec_lo.size, i + 1)

        q11 = buffer[sum(blocks[:-4]): sum(blocks[:-3])].reshape(block_dim)
        q10 = buffer[sum(blocks[:-3]): sum(blocks[:-2])].reshape(block_dim)
        q01 = buffer[sum(blocks[:-2]): sum(blocks[:-1])].reshape(block_dim)
        q00 = buffer[sum(blocks[:-1]): sum(blocks)].reshape(recon_shape)

        quadrants = (q11, q10, q01, q00)

        if i == 0:
            recon = image
        else:
            recon_dim0 = signal_size(block_dim[0], rec_lo.size)
            recon_dim1 = signal_size(block_dim[1], rec_lo.size)
            recon_size = recon_dim0 * recon_dim1
            recon_shape = (recon_dim0, recon_dim1)

            blocks = blocks[:-4]
            blocks.append(recon_size)

            recon = buffer[sum(blocks[:-1]):sum(blocks)].reshape(recon_shape)

        idwt2d_level(recon, quadrants, rec_lo, rec_hi)



import ray
from functools import partial
import numpy as np
import numexpr as ne
import xarray as xr
from pfb.utils.weighting import (_compute_counts, counts_to_weights,
                                 weight_data, filter_extreme_counts)
from pfb.utils.beam import reproject_and_interp_beam
from pfb.utils.fits import set_wcs, save_fits
from pfb.utils.misc import fitcleanbeam
from pfb.operators.gridder import wgridder_conventions
from casacore.quanta import quantity
from datetime import datetime, timezone
from ducc0.fft import good_size
from pfb.utils.astrometry import get_coordinates
from scipy.constants import c as lightspeed
import gc
from astropy import units
from astropy.coordinates import SkyCoord
from africanus.coordinates import radec_to_lm

iFs = np.fft.ifftshift
Fs = np.fft.fftshift

@ray.remote
def batch_stokes_image(
                dc1=None,
                dc2=None,
                operator=None,
                ds=None,
                jones=None,
                opts=None,
                nx=None,
                ny=None,
                freq=None,
                cell_rad=None,
                utime=None,
                rbin_idx=None,
                rbin_counts=None,
                radec=None,
                antpos=None,
                poltype=None,
                fds_store=None,
                bandid=None,
                timeid=None,
                msid=None,
                attrs=None,
                integrations_per_image=None,
                synchronizer=None):
    
    # load chunk
    ds.load(scheduler='sync')
    if jones is not None:
        # we do it this way to force using synchronous scheduler
        jones = jones.load(scheduler='sync').values

    # slice out rows corresponding to single images and submit compute task
    ntime = utime.size
    tasks = []
    # we need to start counting from zero
    rbin_idx = rbin_idx - rbin_idx.min()
    for t0 in range(0, ntime, integrations_per_image):
        nmax = np.minimum(ntime, t0+integrations_per_image)
        It = slice(t0, nmax)
        ridx = rbin_idx[It]
        rcnts = rbin_counts[It]
        Irow = slice(ridx[0], ridx[-1]+rcnts[-1])
        dsi = ds[{'row': Irow}]

        if jones is not None:
            jones_slice = jones[It]
        else:
            jones_slice = None

        task = stokes_image.remote(
                    dc1=dc1,
                    dc2=dc2,
                    operator=operator,
                    ds=dsi,
                    jones=jones_slice,
                    opts=opts,
                    nx=nx,
                    ny=ny,
                    freq=freq,
                    cell_rad=cell_rad,
                    utime=utime[It],
                    tbin_idx=ridx,
                    tbin_counts=rcnts,
                    radec=radec,
                    antpos=antpos,
                    poltype=poltype,
                    fds_store=fds_store,
                    bandid=bandid,
                    timeid=timeid,
                    msid=msid,
                    attrs=attrs)
        
        tasks.append(task)

    # wait for all tasks to finish and get result
    dso = ray.get(tasks)

    # if the output is a dataset we stack and write the output
    if isinstance(dso[0], xr.Dataset):
        dso = xr.concat(dso, dim='TIME')
        dso.to_zarr(fds_store.url,
                    region="auto",
                    synchronizer=synchronizer,
                    safe_chunks=False)
    

    return timeid, bandid
    

@ray.remote
def stokes_image(
                dc1=None,
                dc2=None,
                operator=None,
                ds=None,
                jones=None,
                opts=None,
                nx=None,
                ny=None,
                freq=None,
                cell_rad=None,
                utime=None,
                tbin_idx=None,
                tbin_counts=None,
                radec=None,
                antpos=None,
                poltype=None,
                fds_store=None,
                bandid=None,
                timeid=None,
                msid=None,
                attrs=None):
    # serialization fails for these if we import them above
    from ducc0.misc import resize_thread_pool
    from ducc0.wgridder import vis2dirty
    
    resize_thread_pool(opts.nthreads)
    fieldid = ds.FIELD_ID
    ddid = ds.DATA_DESC_ID
    scanid = ds.SCAN_NUMBER
    oname = f'ms{msid:04d}_fid{fieldid:04d}_spw{ddid:04d}_scan{scanid:04d}' \
            f'_band{bandid:04d}_time{timeid:04d}'

    if opts.precision.lower() == 'single':
        real_type = np.float32
        complex_type = np.complex64
    elif opts.precision.lower() == 'double':
        real_type = np.float64
        complex_type = np.complex128
    
    data = getattr(ds, dc1).values
    ds = ds.drop_vars(dc1)
    if dc2 is not None:
        try:
            assert (operator=='+' or operator=='-')
        except Exception as e:
            raise e
        data = ne.evaluate(f'data {operator} data2',
                            local_dict={'data': data,
                                        'data2': getattr(ds, dc2).values},
                            casting='same_kind')
        ds = ds.drop_vars(dc2)

    time = ds.TIME.values
    # ds = ds.drop_vars('TIME')
    interval = ds.INTERVAL.values
    # ds = ds.drop_vars('INTERVAL')
    ant1 = ds.ANTENNA1.values
    # ds = ds.drop_vars('ANTENNA1')
    ant2 = ds.ANTENNA2.values
    # ds = ds.drop_vars('ANTENNA2')
    uvw = ds.UVW.values
    # ds = ds.drop_vars('UVW')
    flag = ds.FLAG.values
    # ds = ds.drop_vars('FLAG')
    # MS may contain auto-correlations
    frow = ds.FLAG_ROW.values | (ant1 == ant2)
    # ds = ds.drop_vars('FLAG_ROW')

    # combine flag and frow
    flag = np.logical_or(flag, frow[:, None, None])

    nrow, nchan, ncorr = data.shape

    if opts.sigma_column is not None:
        weight = ne.evaluate('1.0/sigma**2',
                             local_dict={'sigma': getattr(ds, opts.sigma_column).values})
        ds = ds.drop_vars(opts.sigma_column)
    elif opts.weight_column is not None:
        weight = getattr(ds, opts.weight_column).values
        ds = ds.drop_vars(opts.weight_column)
    else:
        weight = np.ones((nrow, nchan, ncorr),
                         dtype=real_type)

    if opts.model_column is not None:
        model_vis = getattr(ds, opts.model_column).values.astype(complex_type)
        ds = ds.drop(opts.model_column)
        if opts.product.lower() == 'i':
            model_vis = model_vis[..., [0, -1]].mean(axis=-1, keepdims=True)
        else:
            raise NotImplementedError(f'Model subtraction not supported for product {opts.product}')
    else:
        model_vis = None
    
    # # this seems to help with memory consumption
    # # note the ds.drop_vars above
    # del ds
    # gc.collect()

    nrow, nchan, ncorr = data.shape
    ntime = utime.size
    nant = antpos.shape[0]
    time_out = np.mean(utime)

    freq_out = np.mean(freq)

    if data.dtype != complex_type:
        data = data.astype(complex_type)

    if weight.dtype != real_type:
        weight = weight.astype(real_type)

    if jones is not None:
        if jones.dtype != complex_type:
            jones = jones.astype(complex_type)
        # qcal has chan and ant axes reversed compared to pfb implementation
        jones = np.swapaxes(jones, 1, 2)
        # data are not 2x2 so we need separate labels
        # for jones correlations and data/weight correlations
        # reshape to dispatch with overload
        jones_ncorr = jones.shape[-1]
        if jones_ncorr == 4:
            jones = jones.reshape(ntime, nant, nchan, 1, 2, 2)
        elif jones_ncorr == 2:
            pass
        else:
            raise ValueError("Incorrect number of correlations of "
                            f"{jones_ncorr} for product {opts.product}")
    else:
        jones = np.ones((ntime, nant, nchan, 1, 2),
                        dtype=complex_type)

    # check that there are no missing antennas
    ant1u = np.unique(ant1)
    ant2u = np.unique(ant2)
    allants = np.unique(np.concatenate((ant1u, ant2u)))

    # check that antpos gives the correct size table
    antmax = allants.size
    if opts.check_ants:
        try:
            assert antmax == nant
        except Exception as e:
            raise ValueError('Inconsistent ANTENNA table. '
                            'Shape does not match max number of antennas '
                            'as inferred from ant1 and ant2. '
                            f'Table size is {antpos.shape} but got {antmax}. '
                            f'{oname}')

    # relabel antennas by index
    # this only works because allants is sorted in ascending order
    for a, ant in enumerate(allants):
        ant1 = np.where(ant1==ant, a, ant1)
        ant2 = np.where(ant2==ant, a, ant2)

    cell_deg = np.rad2deg(cell_rad)

    # make sure ra is in (0, 2pi)
    # need a copy since write only
    radec = radec.copy()
    if radec[0] < 0:
        radec[0] += 2*np.pi
    elif radec[0] > 2*np.pi:
        radec[0] -= 2*np.pi

    flip_u, flip_v, flip_w, _, _ = wgridder_conventions(0, 0)
    signu = -1.0 if flip_u else 1.0
    signv = -1.0 if flip_v else 1.0
    # we need these because of flipped wgridder convention
    # https://github.com/mreineck/ducc/issues/34
    signx = -1.0 if flip_u else 1.0
    signy = -1.0 if flip_v else 1.0
    freqfactor = -2j*np.pi*freq[None, :]/lightspeed

    # rephase if asked
    if opts.phase_dir is not None:
        new_ra, new_dec = opts.phase_dir.split(',')
        c = SkyCoord(new_ra, new_dec, frame='fk5', unit=(units.hourangle, units.deg))
        new_ra_rad = np.deg2rad(c.ra.value)
        new_dec_rad = np.deg2rad(c.dec.value)
        radec_new = np.array((new_ra_rad, new_dec_rad))

        # print(c.ra.value, c.dec.value, cell_deg)
        
        from pfb.utils.astrometry import synthesize_uvw
        uvw_new = synthesize_uvw(antpos, time, ant1, ant2, radec_new)
        # uo = uvw[:, 0:1] * freq[None, :]/lightspeed
        # vo = uvw[:, 1:2] * freq[None, :]/lightspeed
        wo = uvw[:, 2:] * freq[None, :]/lightspeed
        # un = uvw_new[:, 0:1] * freq[None, :]/lightspeed
        # vn = uvw_new[:, 1:2] * freq[None, :]/lightspeed
        wn = uvw_new[:, 2:] * freq[None, :]/lightspeed
        
        # TODO - why is this incorrect?
        # from africanus.coordinates import radec_to_lmn
        # l, m, n = radec_to_lmn(radec_new[None, :], radec)[0]
        # # original phase direction is [0,0,1]
        # dl = l
        # dm = m
        # dn = n-1
        # phase2 = uo * dl
        # phase2 += vo * dm
        # phase2 += wo * dn
        # tmp2 = np.exp(-2j*np.pi*phase2)

        # TODO - this copies chgcentre but not sure why it gives
        # better results than computing the phase with lmn differences
        phase = 2j*np.pi*(wn - wo)
        data = data * np.exp(-phase)[:, :, None]
        if model_vis is not None:
            model_vis = model_vis * np.exp(-phase)[:, :, None]
        
        uvw = uvw_new
    else:
        radec_new = radec

    if opts.beam_model is not None:
        # should we compute a weighted mean over freq instead of interpolating here? 
        bds = xr.open_zarr(opts.beam_model, chunks=None).interp(chan=[freq_out])
        l_beam = bds.l_beam.values
        m_beam = bds.m_beam.values
        # the beam is in feed plane direction cosine coordinates
        # we need to flip the beam upside down because of the beam orientation
        # see https://archive-gw-1.kat.ac.za/public/repository/10.48479/wdb0-h061/index.html
        # shape is (corr, chan, X, Y) -> squeeze out freq
        beam = bds.BEAM.values[:, 0, :, :]
        # are the MdV beams transmissive or receptive?
        # reshape for feed and spatial rotations
        beam = beam.reshape(2, 2, l_beam.size, m_beam.size)
        cell_deg_in = l_beam[1] - l_beam[0]
        pbeam = reproject_and_interp_beam(beam, time, antpos,
                                          radec, radec_new,
                                          cell_deg_in, cell_deg, nx, ny,
                                          poltype, opts.product,
                                          weight=weight, nthreads=opts.nthreads)
        
        # this is a hack to get the images to align
        pbeam = np.transpose(pbeam.astype(np.float32),
                             axes=(0, 2, 1))
        pbeam = pbeam[:, ::-1, :]
        
    else:
        pbeam = np.ones((len(opts.product), nx, ny), dtype=real_type)
    
    
    # compute lm coordinates of target if requested
    if opts.target is not None:
        tmp = opts.target.split(',')
        if len(tmp) == 1 and tmp[0] == opts.target:
            obs_time = time_out
            tra, tdec = get_coordinates(obs_time, target=opts.target)
        else:  # we assume a HH:MM:SS,DD:MM:SS format has been passed in
            c = SkyCoord(tmp[0], tmp[1], frame='fk5', unit=(units.hourangle, units.deg))
            tra = np.deg2rad(c.ra.value)
            tdec = np.deg2rad(c.dec.value)

        tcoords=np.zeros((1,2))
        tcoords[0,0] = tra
        tcoords[0,1] = tdec
        lm0 = radec_to_lm(tcoords, radec_new[None, :]).squeeze()
        # flip for wgridder conventions
        x0 = -lm0[0]
        y0 = -lm0[1]
    else:
        x0 = 0.0
        y0 = 0.0
        tra = radec_new[0]
        tdec = radec_new[1]

    ra_deg = np.rad2deg(tra)
    dec_deg = np.rad2deg(tdec)
    # why was this required?
    # if ra_deg > 180:
    #     ra_deg -= 360

    # we currently need this extra loop through the data because
    # we don't have access to the grid
    data, weight = weight_data(
            data, weight, flag, jones,
            tbin_idx, tbin_counts,
            ant1, ant2,
            poltype,
            opts.product,
            str(ncorr))

    # flag if any correlation is flagged
    flag = flag.any(axis=-1)
    mask = (~flag).astype(np.uint8)

    # TODO - this subtraction would be better to do inside weight_data
    if opts.model_column is not None:
        data = ne.evaluate('(data-model_vis)*mask')

    if opts.l2_reweight_dof:
        # data should contain residual_vis at this point
        ressq = (data*data.conj()).real * mask
        wcount = mask.sum()
        if wcount:
            ovar = ressq.sum()/wcount  # use 67% quantile?
            weight = (opts.l2_reweight_dof + 1)/(opts.l2_reweight_dof + ressq/ovar)/ovar
        else:
            weight = None

    
    n = np.sqrt(1 - x0**2 - y0**2)
    psf_vis = np.exp(freqfactor*(signu*uvw[:, 0:1]*x0*signx +
                                 signv*uvw[:, 1:2]*y0*signy -
                                 uvw[:, 2:]*(n-1))).astype(complex_type)

    # TODO - polarisation parameters and handle Stokes axis more elegantly
    # TODO - add beam application to injected transients
    # Should this go before weight_data?
    if opts.inject_transients is not None:
        transient_name = opts.inject_transients.removesuffix('yaml') + 'zarr'
        transient_ds = xr.open_zarr(transient_name, chunks=None)
        all_times = transient_ds.TIME.values
        all_freqs = transient_ds.FREQ.values
        names = transient_ds.names
        ras = transient_ds.ras
        decs = transient_ds.decs
        for name, ra, dec in zip(names, ras, decs):
            ra_rad = np.deg2rad(ra)
            dec_rad = np.deg2rad(dec)
            tcoords=np.zeros((1,2))
            tcoords[0,0] = ra_rad
            tcoords[0,1] = dec_rad
            coords0 = np.array((radec[0], radec[1]))
            lm0t = radec_to_lm(tcoords, coords0).squeeze()
            x0t = lm0t[0]
            y0t = lm0t[1]
            n0t = np.sqrt(1 - x0t**2 - y0t**2)

            # these are the profiles on the full domain so interpolate
            tprofile = getattr(transient_ds, f'{name}_time_profile').values
            fprofile = getattr(transient_ds, f'{name}_freq_profile').values
            
            tprofile = np.interp(time, all_times, tprofile)  # note reorder to ms times
            fprofile = np.interp(freq, all_freqs, fprofile)

            # outer product gives dynamic spectrum
            # LB - why dp we need the braces?
            dspec = (tprofile[:, None]) * fprofile[None, :]

            # inject transient at x0t, y0t and convert to complex values
            dspec = dspec * np.exp(freqfactor*(
                                 signu*uvw[:, 0:1]*x0t*signx +
                                 signv*uvw[:, 1:2]*y0t*signy -
                                 uvw[:, 2:]*(n0t-1)))
            
            # currently Stokes I only
            data[:, :, 0] += dspec

    # TODO - why do we need to cast here?
    data = data.transpose(2, 0, 1).astype(complex_type)
    weight = weight.transpose(2, 0, 1).astype(real_type)
    if opts.robustness is not None:
        # we need to compute the weights on the padded grid
        # but we don't have control over the optimal gridding
        # parameters so assume a minimum
        nx_pad = int(np.ceil(opts.min_padding*nx))
        if nx_pad%2:
            nx_pad += 1
        ny_pad = int(np.ceil(opts.min_padding*ny))
        if ny_pad%2:
            ny_pad += 1
        counts = _compute_counts(uvw,
                                 freq,
                                 mask,
                                 weight,
                                 nx_pad, ny_pad,
                                 cell_rad, cell_rad,
                                 real_type,
                                 k=0,
                                 ngrid=1,
                                 usign=1.0 if flip_u else -1.0,
                                 vsign=1.0 if flip_v else -1.0)

        counts = filter_extreme_counts(counts,
                                       level=opts.filter_counts_level)

        weight = counts_to_weights(
            counts,
            uvw,
            freq,
            weight,
            mask,
            nx_pad, ny_pad,
            cell_rad, cell_rad,
            opts.robustness,
            usign=1.0 if flip_u else -1.0,
            vsign=1.0 if flip_v else -1.0)

    psf_min_size = 128
    if opts.psf_relative_size is None:
        nx_psf = psf_min_size
        ny_psf = psf_min_size
    else:
        nx_psf = good_size(int(nx * opts.psf_relative_size))
        while nx_psf % 2:
            nx_psf = good_size(nx_psf + 1)
        ny_psf = good_size(int(ny * opts.psf_relative_size))
        while ny_psf % 2:
            ny_psf = good_size(ny_psf + 1)
        if nx_psf < psf_min_size or ny_psf < psf_min_size:
            nx_psf = psf_min_size
            ny_psf = psf_min_size

    nstokes = weight.shape[0]
    wsum = np.zeros(nstokes)
    residual = np.zeros((nstokes, nx, ny), dtype=real_type)
    psf = np.zeros((nstokes, nx_psf, ny_psf), dtype=real_type)
    # TODO - the wgridder doesn't check if wgridding is actually
    # required and always makes a minimum of nsupp number of wplanes
    # where nsupp is the gridding kernel support. We could check this
    # with pfb.utils.misc.wplanar if we can figure out the relationship
    # between epsilon on the wplanar threshold parameter. 
    for c in range(nstokes):
        wsum[c] = weight[c, ~flag].sum()
        vis2dirty(
            uvw=uvw,
            freq=freq,
            vis=data[c],
            wgt=weight[c],
            mask=mask,
            npix_x=nx, npix_y=ny,
            pixsize_x=cell_rad, pixsize_y=cell_rad,
            center_x=x0, center_y=y0,
            flip_u=flip_u,
            flip_v=flip_v,
            flip_w=flip_w,
            epsilon=opts.epsilon,
            do_wgridding=opts.do_wgridding,
            divide_by_n=True,  # no rephasing or smooth beam so do it here
            nthreads=opts.nthreads,
            sigma_min=opts.min_padding,
            double_precision_accumulation=opts.double_accum,
            verbosity=0,
            dirty=residual[c])

        vis2dirty(
            uvw=uvw,
            freq=freq,
            vis=psf_vis,
            wgt=weight[c],
            mask=mask,
            npix_x=nx_psf, npix_y=ny_psf,
            pixsize_x=cell_rad, pixsize_y=cell_rad,
            center_x=x0, center_y=y0,
            flip_u=flip_u,
            flip_v=flip_v,
            flip_w=flip_w,
            epsilon=opts.epsilon,
            do_wgridding=opts.do_wgridding,
            divide_by_n=True,  # no rephasing or smooth beam so do it here
            nthreads=opts.nthreads,
            sigma_min=opts.min_padding,
            double_precision_accumulation=opts.double_accum,
            verbosity=0,
            dirty=psf[c])

    # these will be in units of pixels
    GaussPars = fitcleanbeam(psf, level=0.5, pixsize=cell_deg)

    if opts.natural_grad:
        # TODO - add beam application
        from pfb.operators.hessian import hessian_jax
        import jax.numpy as jnp
        from jax.scipy.sparse.linalg import cg
        iFs = jnp.fft.ifftshift

        abspsf = jnp.abs(jnp.fft.rfft2(
                                    iFs(psf/wsum[:, None, None], axes=(1,2)),
                                    axes=(1, 2),
                                    norm='backward'))

        hess = partial(hessian_jax,
                       nx, ny,
                       2*nx, 2*ny,
                       opts.eta,
                       abspsf)

        residual = cg(hess,
               residual/wsum[:, None, None],
               tol=opts.cg_tol,
               maxiter=opts.cg_maxit)[0]
        
    else:
        residual /= wsum[:, None, None]
        residual *= pbeam / (pbeam**2 + opts.eta)

    rms = np.std(residual, axis=(1,2))

    unix_time = quantity(f'{time_out}s').to_unix_time()
    utc = datetime.fromtimestamp(unix_time, tz=timezone.utc).strftime('%Y-%m-%d %H:%M:%S')

    # set corr coords (removing duplicates and sorting)
    corr = "".join(dict.fromkeys(sorted(opts.product)))

    out_ras = ra_deg + np.arange(nx//2, -(nx//2), -1) * cell_deg
    out_decs = dec_deg + np.arange(-(ny//2), ny//2) * cell_deg
    out_ras = np.round(out_ras, decimals=12)
    out_decs = np.round(out_decs, decimals=12)

    # save outputs
    if opts.output_format == 'zarr':
        coords = {
            'FREQ': (('FREQ',), np.array([freq_out])),
            'TIME': (('TIME',), np.mean(utime, keepdims=True)),
            'STOKES': (('STOKES',), list(corr)),
            'X': (('X',), out_ras),
            'Y': (('Y',), out_decs),
        }
        # X and Y are transposed for compatibility with breifast
        data_vars = {}
        residual = np.transpose(residual.astype(np.float32),
                                axes=(0, 2, 1))
        data_vars['cube'] = (('STOKES', 'FREQ', 'TIME', 'Y', 'X'),
                             residual[:, None, None, :, :])
        if opts.psf_out:
            if opts.psf_relative_size == 1:
                xpsf = "X"
                ypsf = "Y"
            else:
                xpsf = "X_PSF"
                ypsf = "Y_PSF"
                coords['X_PSF'] = (('X_PSF',), ra_deg + np.arange(nx_psf//2, -(nx_psf//2), -1) * cell_deg)
                coords['Y_PSF'] = (('Y_PSF',), dec_deg + np.arange(-(ny_psf//2), ny_psf//2) * cell_deg)
            psf /= wsum[:, None, None]
            psf = np.transpose(psf.astype(np.float32),
                               axes=(0, 2, 1))
            data_vars['psf'] = (('STOKES', 'FREQ', 'TIME', ypsf, xpsf),
                                psf[:, None, None, :, :])
        
        if opts.robustness is not None and opts.weight_grid_out:
            ic, ix, iy = np.where(counts > 0)
            wgt = np.zeros_like(counts)
            wgt[ic, ix, iy] = 1.0/counts[ic, ix, iy]
            coords['X_PAD'] = (('X_PAD',), np.arange(nx_pad) * cell_deg)
            coords['Y_PAD'] = (('Y_PAD',), np.arange(ny_pad) * cell_deg)
            wgt = np.transpose(wgt.astype(np.float32),
                               axes=(0, 2, 1))
            data_vars['wgtgrid'] = (('STOKES', 'FREQ', 'TIME', 'Y_PAD', 'X_PAD'),
                                    wgt[:, None, None, :, :])
        
        data_vars['weight'] = (('STOKES','FREQ','TIME'), wsum[:, None, None])
        
        if opts.beam_model is not None:
            weight = pbeam**2 + opts.eta
            weight = np.transpose(weight.astype(np.float32),
                                  axes=(0, 2, 1))
            data_vars['beam_weight'] = (('STOKES', 'FREQ', 'TIME', 'Y', 'X'),
                                    weight[:, None, None, :, :])
        
        data_vars['rms'] = (('STOKES', 'FREQ', 'TIME'), rms[:, None, None].astype(np.float32))
        nonzero = wsum > 0
        data_vars['nonzero'] = (('STOKES', 'FREQ', 'TIME'), nonzero[:, None, None])
        bmaj = np.array([gp[0] for gp in GaussPars], dtype=np.float32)
        bmin = np.array([gp[1] for gp in GaussPars], dtype=np.float32)
        # convert bpa to degrees
        bpa = np.array([gp[2]*180/np.pi for gp in GaussPars], dtype=np.float32)
        data_vars['psf_maj'] = (('STOKES', 'FREQ', 'TIME'), bmaj[:, None, None])
        data_vars['psf_min'] = (('STOKES', 'FREQ', 'TIME'), bmin[:, None, None])
        data_vars['psf_pa'] = (('STOKES', 'FREQ', 'TIME'), bpa[:, None, None])

        if attrs is None:
            attrs = {
                'ra' : tra,
                'dec': tdec,
                'x0': x0,
                'y0': y0,
                'cell_rad': cell_rad,
                'fieldid': fieldid,
                'ddid': ddid,
                'scanid': scanid,
                'bandid': bandid,
                'timeid': timeid,
                'robustness': opts.robustness,
                'utc': utc,
            }

        out_ds = xr.Dataset(
            data_vars,
            coords=coords,
            attrs=attrs)
        if opts.stack:
            return out_ds
        else:
            out_ds.to_zarr(f'{fds_store.url}/{oname}.zarr', mode='w')
    elif opts.output_format == 'fits':
        # if there is more than one polarisation product
        # we currently assume all have the same beam
        hdr = set_wcs(cell_deg, cell_deg, nx, ny, [tra, tdec],
                    freq_out, GuassPar=GaussPars[0],
                    ms_time=time_out, ncorr=len(corr))

        hdr['STOKES'] = corr
        save_fits(residual/wsum[:, None, None],
                  f'{fds_store.full_path}/{oname}_image.fits', hdr)
        if opts.robustness is not None and opts.weight_grid_out:
            ic, ix, iy = np.where(counts > 0)
            wgt = np.zeros_like(counts)
            wgt[ic, ix, iy] = 1.0/counts[ic, ix, iy]
            # TODO - add mirror image
            save_fits(wgt,
                    f'{fds_store.full_path}/{oname}_weight.fits', hdr)

        if opts.psf_out:
            hdr_psf = set_wcs(cell_deg, cell_deg, nx_psf, ny_psf, [tra, tdec],
                  freq_out, GuassPar=GaussPars[0],  # fake for now
                  ms_time=time_out)
            hdr_psf['STOKES'] = corr
            save_fits(psf/wsum[:, None, None],
                  f'{fds_store.full_path}/{oname}_psf.fits', hdr_psf)

        if opts.beam_model is not None:
            # weight = np.transpose(weight.astype(np.float32),
            #                      axes=(0, 2, 1))
            # weight = weight[:, ::-1, :]
            weight = pbeam**2 + opts.eta
            save_fits(weight,
                  f'{fds_store.full_path}/{oname}_weight.fits', hdr)
    return 1



import numpy as np
from astropy.io import fits
from astropy.wcs import WCS
from datetime import datetime, timezone
from casacore.quanta import quantity
from astropy.time import Time
from pfb.utils.naming import xds_from_list
import xarray as xr


def to4d(data):
    if data.ndim == 4:
        return data
    elif data.ndim == 2:
        return data[None, None]
    elif data.ndim == 3:
        return data[None]
    elif data.ndim == 1:
        return data[None, None, None]
    else:
        raise ValueError("Only arrays with ndim <= 4 can be broadcast to 4D.")


def data_from_header(hdr, axis=3):
    npix = hdr['NAXIS' + str(axis)]
    refpix = hdr['CRPIX' + str(axis)]
    delta = hdr['CDELT' + str(axis)]
    ref_val = hdr['CRVAL' + str(axis)]
    return ref_val + np.arange(1 - refpix, 1 + npix - refpix) * delta, ref_val


def load_fits(name, dtype=np.float32):
    data = fits.getdata(name)
    data = np.transpose(to4d(data), axes=(1, 0, 3, 2))  # fits and beams table
    return np.require(data, dtype=dtype, requirements='C')


def save_fits(data, name, hdr, overwrite=True, dtype=np.float32, beams_hdu=None):
    hdu = fits.PrimaryHDU(header=hdr)
    data = np.transpose(to4d(data), axes=(1, 0, 3, 2))
    hdu.data = np.require(data, dtype=dtype, requirements='F')
    if beams_hdu is not None:
        hdul = fits.HDUList([hdu, beams_hdu])
        hdul.writeto(name, overwrite=overwrite)
    else:
        hdu.writeto(name, overwrite=overwrite)
    return


def set_wcs(cell_x, cell_y, nx, ny, radec, freq,
            unit='Jy/beam', GuassPar=None, ms_time=None,
            header=True, casambm=True, ncorr=1):
    """
    cell_x/y - cell sizes in degrees
    nx/y - number of x and y pixels
    radec - right ascention and declination in radians
    freq - frequencies in Hz
    unit - Jy/beam or Jy/pixel
    GuassPar - MFS beam parameters in degrees
    ms_time - measurement set time
    header - if True, return a header, otherwise return a WCS object
    casambm - if True, add the CASAMBM keyword to the header
    """

    w = WCS(naxis=4)
    w.wcs.ctype = ['RA---SIN', 'DEC--SIN', 'FREQ', 'STOKES']
    w.wcs.cdelt[0] = -cell_x
    w.wcs.cdelt[1] = cell_y
    w.wcs.cdelt[3] = 1
    w.wcs.cunit[0] = 'deg'
    w.wcs.cunit[1] = 'deg'
    w.wcs.cunit[2] = 'Hz'
    w.wcs.cunit[3] = ''
    if np.size(freq) > 1:
        nchan = freq.size
        crpix3 = nchan//2+1
        ref_freq = freq[crpix3]
        df = freq[1]-freq[0]
        w.wcs.cdelt[2] = df
    else:
        if isinstance(freq, np.ndarray) and freq.size == 1:
            ref_freq = freq[0]
        else:
            ref_freq = freq
        nchan = 1
        crpix3 = 1
    w.wcs.crval = [radec[0]*180.0/np.pi, radec[1]*180.0/np.pi, ref_freq, 1]
    w.wcs.crpix = [1 + nx//2, 1 + ny//2, crpix3, 1]
    w.wcs.equinox = 2000.0

    if header:
        # the order does seem to matter here,
        # especially when using with StreamingHDU
        header = fits.Header()
        header['SIMPLE'] = (True, 'conforms to FITS standard')
        header['BITPIX'] = (-32, 'array data type')
        header['NAXIS'] = (4, 'number of array dimensions')
        data_shape = (nx, ny, nchan, ncorr)
        for i, size in enumerate(data_shape, 1):
            header[f'NAXIS{i}'] = (size, f'length of data axis {i}')
        
        header['EXTEND'] = True
        header['BSCALE'] = 1.0
        header['BZERO'] = 0.0
        header['BUNIT'] = unit
        header['EQUINOX'] = 2000.0
        header['BTYPE'] = 'Intensity'
        

        # add wcs keywords
        wcs_header = w.to_header()
        header.update(wcs_header)

        header['RESTFRQ'] = ref_freq
        header['ORIGIN'] = 'pfb-imaging'
        header['SPECSYS'] = 'TOPOCENT'
        if ms_time is not None:
            # TODO - probably a round about way of doing this
            unix_time = quantity(f'{ms_time}s').to_unix_time()
            utc_iso = datetime.fromtimestamp(unix_time,
                            tz=timezone.utc).strftime('%Y-%m-%d %H:%M:%S')
            header['UTC_TIME'] = utc_iso
            t = Time(utc_iso)
            t.format = 'fits'
            header['DATE-OBS'] = t.value

        # What are these used for?
        # if 'LONPOLE' in header:
        #     header.pop('LONPOLE')
        # if 'LATPOLE' in header:
        #     header.pop('LATPOLE')
        # if 'RADESYS' in header:
        #     header.pop('RADESYS')
        # if 'MJDREF' in header:
        #     header.pop('MJDREF')

        # 
        
        if casambm:
            header['CASAMBM'] = casambm  # we need this to pick up the beams table
        
        if GuassPar is not None:
            header = add_beampars(header, GuassPar)

        return header
    else:
        return w


def add_beampars(hdr, GaussPar, GaussPars=None, unit2deg=1.0):
    """
    Add beam keywords to header.
    GaussPar - MFS beam pars
    GaussPars - beam pars for cube
    unit2deg - conversion factor to convert BMAJ/BMIN to degrees

    PA is passed in radians and follows the parametrisation in
    
    pfb/utils/misc/Gaussian2D
    
    """
    if len(GaussPar) == 1:
        GaussPar = GaussPar[0]
    elif len(GaussPar) != 3:
        raise ValueError('Invalid value for GaussPar')

    if not np.isnan(GaussPar).any():
        hdr['BMAJ'] = GaussPar[0]*unit2deg
        hdr['BMIN'] = GaussPar[1]*unit2deg
        hdr['BPA'] = GaussPar[2]*180/np.pi

    if GaussPars is not None:
        for i in range(len(GaussPars)):
            if not np.isnan(GaussPars[i]).any():
                hdr['BMAJ' + str(i+1)] = GaussPars[i][0]*unit2deg
                hdr['BMIN' + str(i+1)] = GaussPars[i][1]*unit2deg
                hdr['BPA' + str(i+1)] = GaussPars[i][2]*180/np.pi

    return hdr


def create_beams_table(beams_data, cell2deg):
    """
    Add a BEAMS subtable to a FITS file.
    
    Parameters:
    -----------
    filename : str
        The FITS file to modify
    beams_data : dict
        Dictionary containing arrays for:
        - chan_id: Channel indices
        - pol_id: Polarization indices
        - bmaj: Major axis values (in pixels)
        - bmin: Minor axis values (in pixels)
        - bpa: Position angles (in radians)
    """
    # Create the columns for the BEAMS table
    nband = beams_data.band.size
    npol = beams_data.corr.size
    band_id = []
    pol_id = []
    for b in range(nband):
        for p in range(npol):
            band_id.append(b)
            pol_id.append(p)

    # we need the transpose for C -> F ordering
    bmaj = beams_data.sel({'bpar': 'BMAJ'}).values.ravel() * cell2deg
    bmin = beams_data.sel({'bpar': 'BMIN'}).values.ravel() * cell2deg
    bpa = beams_data.sel({'bpar': 'BPA'}).values.ravel() * 180/np.pi
    col1 = fits.Column(name='BMAJ', format='1E', array=bmaj, unit='deg')
    col2 = fits.Column(name='BMIN', format='1E', array=bmin, unit='deg')
    col3 = fits.Column(name='BPA', format='1E', array=bpa, unit='deg')
    col4 = fits.Column(name='CHAN', format='1J', array=np.array(band_id))
    col5 = fits.Column(name='POL', format='1J', array=np.array(pol_id))
    
    # Create the BEAMS table HDU
    cols = fits.ColDefs([col1, col2, col3, col4, col5])
    beams_hdu = fits.BinTableHDU.from_columns(cols)
    beams_hdu.name = 'BEAMS'
    beams_hdu.header['EXTNAME'] = 'BEAMS'
    beams_hdu.header['EXTVER'] = 1
    beams_hdu.header['XTENSION'] = 'BINTABLE'
    beams_hdu.header.comments['XTENSION'] = 'Binary extension'
    beams_hdu.header['NCHAN'] = nband
    beams_hdu.header['NPOL'] = npol

    return beams_hdu


def dds2fits(dsl, column, outname, norm_wsum=True,
             otype=np.float32, nthreads=1,
             do_mfs=True, do_cube=True,
             psfpars_mfs=None):
    basename = outname + '_' + column.lower()
    if norm_wsum:
        unit = 'Jy/beam'
    else:
        unit = 'Jy/pixel'
    dds = xds_from_list(dsl, drop_all_but=[column, 'PSFPARSN', 'WSUM'],
                        nthreads=nthreads,
                        order_freq=False)
    timeids = np.unique(np.array([int(ds.timeid) for ds in dds]))
    freqs = [ds.freq_out for ds in dds]
    freqs = np.unique(freqs)
    nband = freqs.size
    for timeid in timeids:
        # filter by time ID
        dst = [ds for ds in dds if int(ds.timeid)==timeid]
        # concat creating a new band axis
        dsb = xr.concat(dst, dim='band')
        # LB - these seem to remain in the correct order
        dsb = dsb.assign_coords({'band': np.arange(nband)})
        wsums = dsb.WSUM.values
        wsum = dsb.WSUM.sum(dim='band').values
        cube = dsb.get(column).values
        _, _, nx, ny = cube.shape
        # these should be the same across band and corr axes
        radec = (dsb.ra, dsb.dec)
        cell_deg = np.rad2deg(dsb.cell_rad)

        if do_mfs:
            # we need a single freq_mfs for the cube
            freq_mfs = np.sum(freqs[:, None]*wsums)/wsum.sum()
            if norm_wsum:
                # already weighted by wsum
                cube_mfs = np.sum(cube, axis=0)/wsum[:, None, None]
            else:
                # weigted sum
                cube_mfs = np.sum(cube*wsums[:, :, None, None],
                                  axis=0)/wsum[:, None, None]

            name = basename + f'_time{dsb.timeid}_mfs.fits'
            hdr = set_wcs(cell_deg, cell_deg, nx, ny, radec, freq_mfs,
                          unit=unit, ms_time=dsb.time_out)
            # hdr['WSUM'] = wsum
            if psfpars_mfs is not None:
                da_mfs = xr.DataArray(data=np.array(psfpars_mfs[dsb.timeid])[None, :, :],
                                      coords={'band': np.arange(1),
                                              'corr': dsb.corr.values,
                                              'bpar': dsb.bpar.values})
                beams_hdu = create_beams_table(da_mfs, cell2deg=cell_deg)
            else:
                beams_hdu = None

            save_fits(cube_mfs, name, hdr, overwrite=True,
                      dtype=otype, beams_hdu=beams_hdu)

        if do_cube:
            if norm_wsum:
                cube = cube/wsums[:, :, None, None]
            name = basename + f'_time{dsb.timeid}.fits'
            hdr = set_wcs(cell_deg, cell_deg, nx, ny, radec, freqs,
                          unit=unit, ms_time=dsb.time_out)
            
            if 'PSFPARSN' in dsb:
                beams_hdu = create_beams_table(dsb.PSFPARSN, cell2deg=cell_deg)
            save_fits(cube, name, hdr, overwrite=True,
                      dtype=otype, beams_hdu=beams_hdu)

    return column




'''
This is an attempt to develop a minor cycle for the resolve polarisation model.

The rationale here is that primary beam patterns are actually for corellations, not Stokes parameters.

As a starting point, we try to figure out how to compute the dirty image as
a convolution of the model with the PSF i.e.

ID = PSF.conv(B) = R.H Ninv V

where B is the brightness matrix.

'''


import numpy as np
import jax
jax.config.update("jax_enable_x64", True)
import jax.numpy as jnp
from jax.scipy.linalg import expm
from jax._src.scipy.sparse.linalg import _vdot_tree
import nifty8.re as jft
from functools import partial
from ducc0.wgridder import dirty2vis, vis2dirty
from ducc0.fft import c2c, r2c, c2r
from ducc0.nufft import u2nu, nu2u
from numpy.fft import fftshift, ifftshift
Fs = fftshift
iFs = ifftshift
import matplotlib.pyplot as plt
from pfb.utils.stokes import stokes_to_corr, corr_to_stokes


@jax.jit
def pol_energy_approx(ID, psfhat, x):
    """
    x - (4, nx, nx) - signal
    ID - (nx, nx, 4) - dirty image
    psfhat - (nx_psf, ny_psf, 4) FFT of point spread function
    """
    psfh = jax.lax.stop_gradient(psfhat)
    Id = jax.lax.stop_gradient(ID)
    nx_psf, ny_psf, _ = psfh.shape
    nx, ny, _ = Id.shape
    # Convert to brightness
    a = x[0:n*n].reshape(n, n)
    b = x[n*n:2*n*n].reshape(n, n)
    c = x[2*n*n:3*n*n].reshape(n, n)
    d = x[3*n*n:].reshape(n, n)
    y = jnp.stack((a+b, c+1j*d, c-1j*d, a-b), axis=-1).reshape(nx, ny, 2, 2)
    B = expm(y).reshape(nx, ny, 4)
    # Convolution
    Bhat = jnp.fft.fft2(B,
                        s=(nx_psf, ny_psf),
                        axes=(0, 1),
                        norm='backward')
    Bconv = jnp.fft.ifft2(Bhat * psfh,
                         s=(nx_psf, ny_psf),
                         axes=(0, 1),
                         norm='backward')[0:nx, 0:ny, :]
    # # _vdot_tree should promote to double during vdot
    # return (_vdot_tree(B, Bconv).real -
    #         _vdot_tree(B, Id).real*2 +
    #         _vdot_tree(a, a) + 
    #         _vdot_tree(b, b) +
    #         _vdot_tree(c, c) +
    #         _vdot_tree(d, d))
    
    return (jnp.vdot(B, Bconv).real -
            jnp.vdot(B, Id).real*2)
            # jnp.sum(a*a) + 
            # jnp.sum(b*b) +
            # jnp.sum(c*c) +
            # jnp.sum(d*d))


def stokes_from_abcd(a, b, c, d):
    p = np.sqrt(b**2 + c**2 + d**2)
    I = np.exp(a)*np.cosh(p)
    Q = b*np.exp(a)*np.sinh(p)/p
    U = c*np.exp(a)*np.sinh(p)/p
    V = d*np.exp(a)*np.sinh(p)/p
    return I, Q, U, V


if __name__=="__main__":
    n = 128
    I = np.zeros((n, n))
    Q = np.zeros((n, n))
    U = np.zeros((n, n))
    V = np.zeros((n, n))

    np.random.seed(42)
    idx = np.random.randint(10, n-10, 10)
    idy = np.random.randint(10, n-10, 10)

    Q[idx, idy] = 1
    U[idx, idy] = -1
    V[idx, idy] = 0.1
    p = 0.7
    I = p*np.sqrt(Q**2 + U**2 + V**2)/p

    # make vis
    m = 10000
    u = np.random.uniform(-1, 1, m)
    v = np.random.uniform(-1, 1, m)
    w = np.zeros(m)
    uvw = np.stack((u, v, w), axis=-1)

    freq = np.array([1e9])

    lightspeed = 299792458.0
    uv_max = 2.0  # in (-1, 1)
    cell_N = 1.0 / (2 * uv_max * freq[0] / lightspeed)
    epsilon = 1e-6  # gridding precision

    # convert Stokes to vis
    vis_I = dirty2vis(
        uvw=uvw,
        freq=freq,
        dirty=I,
        pixsize_x=cell_N,
        pixsize_y=cell_N,
        epsilon=epsilon,
    )

    vis_Q = dirty2vis(
        uvw=uvw,
        freq=freq,
        dirty=Q,
        pixsize_x=cell_N,
        pixsize_y=cell_N,
        epsilon=epsilon,
    )

    vis_U = dirty2vis(
        uvw=uvw,
        freq=freq,
        dirty=U,
        pixsize_x=cell_N,
        pixsize_y=cell_N,
        epsilon=epsilon,
    )

    vis_V = dirty2vis(
        uvw=uvw,
        freq=freq,
        dirty=V,
        pixsize_x=cell_N,
        pixsize_y=cell_N,
        epsilon=epsilon,
    )

    # convert Stokes vis to corr
    vis = np.zeros((m, 1, 4), dtype=np.complex128)
    vis[:, :, 0] = vis_I + vis_Q
    vis[:, :, 1] = vis_U + 1j*vis_V
    vis[:, :, 2] = vis_U - 1j*vis_V
    vis[:, :, 3] = vis_I - vis_Q


    # create dirty images (only off diagonal elements are expected to be complex)
    dirty0 = vis2dirty(uvw=uvw, freq=freq, vis=vis[:, :, 0],
                       npix_x=n, npix_y=n,
                       pixsize_x=cell_N, pixsize_y=cell_N,
                       epsilon=epsilon, do_wgridding=False,
                       nthreads=8).astype("f8")
    dirty1 = vis2dirty(uvw=uvw, freq=freq, vis=vis[:, :, 1],
                       npix_x=n, npix_y=n,
                       pixsize_x=cell_N, pixsize_y=cell_N,
                       epsilon=epsilon, do_wgridding=False,
                       nthreads=8).astype("f8") \
                +1j * vis2dirty(uvw=uvw, freq=freq, vis=-1j*vis[:, :, 1],
                                npix_x=n, npix_y=n,
                                pixsize_x=cell_N, pixsize_y=cell_N,
                                epsilon=epsilon, do_wgridding=False,
                                nthreads=8).astype("f8")
    dirty2 = vis2dirty(uvw=uvw, freq=freq, vis=vis[:, :, 2],
                       npix_x=n, npix_y=n,
                       pixsize_x=cell_N, pixsize_y=cell_N,
                       epsilon=epsilon, do_wgridding=False,
                       nthreads=8).astype("f8") \
                +1j * vis2dirty(uvw=uvw, freq=freq, vis=-1j*vis[:, :, 2],
                                npix_x=n, npix_y=n,
                                pixsize_x=cell_N, pixsize_y=cell_N,
                                epsilon=epsilon, do_wgridding=False,
                                nthreads=8).astype("f8")
    dirty3 = vis2dirty(uvw=uvw, freq=freq, vis=vis[:, :, 3],
                       npix_x=n, npix_y=n,
                       pixsize_x=cell_N, pixsize_y=cell_N,
                       epsilon=epsilon, do_wgridding=False,
                       nthreads=8).astype("f8")
    
    ID = np.stack((dirty0, dirty1, dirty2, dirty3), axis=-1)

    # in the absence of gains the PSF for each correlation should be the same
    psf = vis2dirty(uvw=uvw, freq=freq, vis=2*np.ones((m, 1), dtype=np.complex128),
                    npix_x=2*n, npix_y=2*n,
                    pixsize_x=cell_N, pixsize_y=cell_N,
                    epsilon=epsilon, do_wgridding=False,
                    nthreads=8).astype("f8")
    
    wsum = psf.max()

    dirtyI, dirtyQ, dirtyU, dirtyV = corr_to_stokes(ID, wsum=wsum)


    plt.figure('1')
    plt.imshow(dirtyI)
    plt.colorbar()

    plt.figure('2')
    plt.imshow(dirtyQ)
    plt.colorbar()

    plt.figure('3')
    plt.imshow(dirtyU)
    plt.colorbar()

    plt.figure('4')
    plt.imshow(dirtyV)
    plt.colorbar()

    plt.show()


    # don't use rfft since corrs could be complex
    psfhat = np.fft.fft2(iFs(psf), axes=(0, 1), norm='backward')
    psfhat = np.tile(psfhat[:, :, None], (1, 1, 4))


    # test convolve
    S = np.stack([I, Q, U, V], axis=-1)
    B = np.stack(stokes_to_corr(S), axis=-1)
    # Convolution
    Bhat = jnp.fft.fft2(B,
                        s=(2*n, 2*n),
                        axes=(0, 1),
                        norm='backward')
    Bconv = jnp.fft.ifft2(Bhat * psfhat,
                          s=(2*n, 2*n),
                          axes=(0, 1),
                          norm='backward')[0:n, 0:n, :]/2
    
    # convert back to stokes
    dirtyI2, dirtyQ2, dirtyU2, dirtyV2 = corr_to_stokes(Bconv, wsum=wsum)

    # compare Stokes dirty images
    resI = np.abs(dirtyI - dirtyI2)
    resQ = np.abs(dirtyQ - dirtyQ2)
    resU = np.abs(dirtyU - dirtyU2)
    resV = np.abs(dirtyV - dirtyV2)

    assert np.allclose(resI, 0, atol=epsilon)
    assert np.allclose(resQ, 0, atol=epsilon)
    assert np.allclose(resU, 0, atol=epsilon)
    assert np.allclose(resV, 0, atol=epsilon)

    # compare corr dirty images
    res0 = ID[:, :, 0] - Bconv[:, :, 0]
    res1 = ID[:, :, 1] - Bconv[:, :, 1]
    res2 = ID[:, :, 2] - Bconv[:, :, 2]
    res3 = ID[:, :, 3] - Bconv[:, :, 3]



import numpy as np
from numba import njit, prange


def prox_21(v, sigma, weight=None, axis=0):
    """
    Computes weighted version of

    prox_{sigma * || . ||_{21}}(v)

    Assumed that v has shape (nband, nbasis, ntot) where

    nband   - number of imaging bands
    nbasis  - number of orthogonal bases
    ntot    - total number of coefficients for each basis (must be equal)
    """
    l2_norm = np.linalg.norm(v, axis=axis)  # drops axis
    l2_soft = np.maximum(l2_norm - sigma * weight, 0.0)  # norm positive
    mask = l2_norm != 0
    ratio = np.zeros(mask.shape, dtype=v.dtype)
    ratio[mask] = l2_soft[mask] / l2_norm[mask]
    return v * np.expand_dims(ratio, axis=axis)  # restores axis


@njit(nogil=True, cache=True, parallel=True)
def prox_21_numba(v, result, lam, sigma=1.0, weight=None):
    """
    Computes weighted version of

    prox_{sigma * || . ||_{21}}(v)

    Assumed that v has shape (nband, nbasis, ntot) where

    nband   - number of imaging bands
    nbasis  - number of orthogonal bases
    ntot    - total number of coefficients for each basis (must be equal)
    """
    nband, nbasis, ntot = v.shape
    # sum over band axis upfront?
    # vsum = np.sum(v, axis=0)
    # result = np.zeros((nband, nbasis, ntot))
    for b in range(nbasis):
        vb = v[:, b]
        weightb = weight[b]
        resultb = result[:, b]
        for i in prange(ntot):
            absvbi = np.linalg.norm(vb[:, i])/sigma
            if not absvbi:
                resultb[:, i] = 0.0
                continue
            softvbi = np.maximum(absvbi - lam*weightb[i]/sigma, 0.0) #* vbisum/absvbi
            resultb[:, i] = vb[:, i] * softvbi / absvbi /sigma


def dual_update(v, x, psiH, lam, sigma=1.0, weight=1.0):
    vp = v.copy()
    vout = np.zeros_like(v)
    psiH(x, vout)
    vtilde = vp + sigma * vout
    # return vtilde
    v = vtilde - sigma * prox_21(vtilde/sigma, lam/sigma, weight=weight)
    return v



@njit(nogil=True, cache=True, parallel=True)
def dual_update_numba(vp, v, lam, sigma=1.0, weight=None):
    """
    Computes weighted version of

    prox_{sigma * || . ||_{21}}(vp)

    Assumed that v has shape (nband, nbasis, ntot) where

    nband   - number of imaging bands
    nbasis  - number of orthogonal bases
    ntot    - total number of coefficients for each basis (must be equal)

    v is initialised with psiH(xp) and will be updated
    """
    nband, nbasis, ntot = v.shape
    for b in range(nbasis):
        vtildeb = vp[:, b, :] + sigma * v[:, b, :]
        weightb = weight[b]
        for i in prange(ntot):
            vtildebi = vtildeb[:, i]
            absvbisum = np.linalg.norm(vtildebi)/sigma
            v[:, b, i] = vtildebi
            if absvbisum:
                softvbi = np.maximum(absvbisum - lam*weightb[i]/sigma, 0.0)
                v[:, b, i] *= (1-softvbi / absvbisum)
